"""This class pop a tool view form in a subdialog"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.application.dialogs.ChildDialogView import ChildDialogView

class DummyMainApp:
    def __init__(self, settings):
        self.settings = settings

class ChildDialogGenericView(ChildDialogView):
    """
    Open a child dialog of a tkinter application to answer a question.
    """
    def __init__(self, parent, title, view, is_insert=False):
        """
        Open a child dialog of a tkinter application to choose autoscan settings.

        Args:
            parent: the tkinter parent view to use for this window construction.
            title: title of the window
            view : A view object to display
            is_insert(False) : open in insert mode
        """
        super().__init__(parent, title)
        self.view = view
        self.view.appliViewFrame = self.appFrame
        self.viewdelete = self.view.delete
        self.viewinsert = self.view.insert
        self.view.delete = self.deleteProxy
        self.view.insert = self.insertProxy
        if is_insert:
            self.view.openInsertWindow()
        else:
            self.view.openModifyWindow()
        self.completeDialogView(addButtons=False)

    def deleteProxy(self, *args, **kwargs):
        self.viewdelete(*args, **kwargs)
        self.app.destroy()
   
    def insertProxy(self, *args, **kwargs):
        res = self.viewinsert(*args, **kwargs)
        self.rvalue = res
        self.app.destroy()

  


"""Controller for Tool object. Mostly handles conversion between mongo data and python objects"""
from pollenisatorgui.core.controllers.controllerelement import ControllerElement


class ToolController(ControllerElement):
    """Inherits ControllerElement
    Controller for Tool object. Mostly handles conversion between mongo data and python objects"""

    def doUpdate(self, values):
        """
        Update the Tool represented by this model in database with the given values.

        Args:
            values: A dictionary crafted by ToolView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated Tool document.
        """
        # Get form values
        dated = values.get("Start date", self.model.dated)
        datef = values.get("End date", self.model.datef)
        self.model.scanner_ip = values.get("Scanner", self.model.scanner_ip)
        self.model.notes = values.get("Notes", self.model.notes)
        # update fetched instance before update, in case there's a Hook after update.
        self.model.dated = dated
        self.model.datef = datef
        # Updating
        self.update()

    def getStatus(self):
        """Returns a string describing the tool current status
        Returns:
            string with possible values : "OOS"/"OOT"/"running"/"done". OOS = Out of Scope, OOT = Out of Time range
        """
        return self.model.getStatus()
        
    def setStatus(self, status):
        """Set the tool model status
        Args:
            status: string with possible values : "OOS"/"OOT"/"running"/"done". OOS = Out of Scope, OOT = Out of Time range
        """
        self.model.setStatus(status)

    def getName(self):
        """Returns the model tool name
        Returns: 
            string
        """
        return self.model.name

    
    def getOutputDir(self, pentestName):
        """Returns directory of the tool file output 
        Args:
            pentestName: the pentest database name
        Returns:
            string (path)
        """
        return self.model.getOutputDir(pentestName)

    def getResultFile(self):
        """Returns path of the tool resulting file output
        Returns:
            string (path)
        """
        return self.model.getResultFile()

    def markAsNotDone(self):
        """Change this model tool to status not done. (resets dates and scanner)
        """
        self.model.markAsNotDone()

    def getType(self):
        """Returns a string describing the type of object
        Returns:
            "tool" """
        return "tool"


"""Frame for remote terminal emulation"""
import tkinter as tk
import tkinter.ttk as ttk
import traceback
from customtkinter import *
from pollenisatorgui.core.components.apiclient import APIClient
import threading
import pollenisatorgui.core.components.utils as utils


class PseudoTermFrame(CTkFrame):
    """
    Open a child dialog of a tkinter application to show a text area.
    """

    def __init__(self, parent, toolController, scanManager):
        """
        Open a child dialog of a tkinter application to ask details about
        existing files parsing.

        Args:
            default_path: a default path to be added
        """
        super().__init__(parent)
        self.parent = parent
        self.toolController = toolController
        self.rvalue = None
        self.text_area = CTkTextbox(self, font=("Consolas", 14), fg_color="#000000", text_color="#00ff00")
        self.text_area.insert(tk.END, toolController.getDetailedString()+"\n"+str(toolController.getData().get("infos",{}).get("cmdline", ""))+"\n")
        self.text_area.pack(expand=True, fill=tk.BOTH)     
        self.text_area.focus_force()
        self.pack(expand=True)
        self.sm = scanManager
        self.timer = threading.Timer(0.5, self.getProgress)
        self.timer.start()
            
    
    def getProgress(self):
        # Print the key that was pressed
        if self.sm.is_local_launched(str(self.toolController.getDbId())):
            result = self.sm.getToolProgress(str(self.toolController.getDbId()))
        else:
            apiclient = APIClient.getInstance()
            result = apiclient.getToolProgress(str(self.toolController.getDbId()))
        try:
            if isinstance(result, str):
                self.text_area.insert(tk.END, result.replace("\r","\n"))
            elif isinstance(result, bool) and result:
                self.text_area.insert(tk.END, "Scan ended. You can quit and download result file.")
                return True
            elif isinstance(result, bool) and not result:
                pass #
            elif isinstance(result, bytes):
                self.text_area.insert(tk.END, result.decode("utf-8").replace("\r","\n"))
            elif result is None:
                return True
            else:
                self.text_area.insert(tk.END, f"Could not get progress : {result[1]}")
                return False
        except Exception as e:
            self.text_area.insert(tk.END, f"Could not get progress : {e}")
            traceback.print_exc()
            self.timer.cancel()
            return False
        self.timer = threading.Timer(0.2, self.getProgress)
        self.timer.start()
        #self.sio.emit("remoteInteractionGet", {"name":apiclient.getUser(), "db": apiclient.getCurrentPentest(),"id":str(self.toolController.getDbId())})
        return True

    def quit(self):
        if self.timer is not None:
            self.timer.cancel()
            


"""Widget with no display that holds a value"""
from pollenisatorgui.core.forms.form import Form
from tkinter import ttk
import tkinter as tk 

class FormSeparator(Form):
    """
    Form field hidden, to store a value.
    """

    def __init__(self, **kwargs):
        """
        Constructor for a hidden form.

        Args:
            name: the form name.
            default: a default value to store in it.
        """
        super().__init__("Separator")
        self.kwargs = kwargs


    def constructView(self, parent):
        """
        Create the button view inside the parent view given

        Args:
            parent: parent form panel.
        """
        self.sep = ttk.Separator(parent.panel,  orient=self.getKw("orient", "horizontal"))
        if parent.gridLayout:
            self.sep.grid(row=self.getKw("row", 0), column=self.getKw("column", 0), columnspan=self.getKw("columnspan", 1), sticky=self.getKw("sticky", tk.W), **self.kwargs)
        else:
            self.sep.pack(side=self.getKw("side", "top"), padx=self.getKw("padx", 5), pady=self.getKw("pady", 5), fill=self.getKw("fill", tk.X),**self.kwargs)




class TagInfos:
    def __init__(self, *args, **kwargs):
        if isinstance(args[0], tuple) or isinstance(args[0], list):
            name = args[0][0]
            color = args[0][1]
            level = args[0][2]
            try:
                notes = args[0][3]
            except IndexError:
                notes = None
        elif isinstance(args[0], TagInfos):
            name = args[0].name
            color = args[0].color
            level = args[0].level
            notes = args[0].notes
        elif isinstance(args[0], dict):
            name = args[0]["name"]
            color = args[0]["color"]
            level = args[0]["level"]
            notes = args[0].get("notes", "")
        elif isinstance(args[0], str):
            name = args[0]
            color = kwargs.get("color", None)
            level = kwargs.get("level", None)
            notes = kwargs.get("notes", None)
        else:
            raise TypeError("Tag constructor only accepts dict, tuple, TagInfos or str and got "+str(type(args[0])))
        if kwargs.get("color", None) is not None:
            color = kwargs.get("color", None)
        if kwargs.get("level", None) is not None:
            level = kwargs.get("level", None)
        if kwargs.get("notes", None) is not None:
            notes = kwargs.get("notes", None)
        self.name = name
        self.color = color if color is not None else "transparent"
        self.level = level if level is not None else "info"
        self.notes = notes if notes is not None else ""

    def getData(self):
        return {"name": self.name, "color": self.color, "level": self.level, "notes": self.notes}


"""Controller for Wave object. Mostly handles conversion between mongo data and python objects"""
from pollenisatorgui.core.controllers.controllerelement import ControllerElement


class WaveController(ControllerElement):
    """Inherits ControllerElement
    Controller for Wave object. Mostly handles conversion between mongo data and python objects"""

    def doUpdate(self, values):
        """
        Update the Wave represented by this model in database with the given values.

        Args:
            values: A dictionary crafted by WaveView containg all form fields values needed.

        Returns:
            The mongo ObjectId _id of the updated Wave document.
        """
        wave_commands_dict = values.get("Commands", self.model.wave_commands)
        wave_commands = [k for k, v in wave_commands_dict.items() if v == 1]
        
        self.model.wave_commands = list(wave_commands)
        self.model.update()
        

    def doInsert(self, values):
        """
        Insert the Wave represented by this model in the database with the given values.

        Args:
            values: A dictionary crafted by WaveView containing all form fields values needed.

        Returns:
            {
                '_id': The mongo ObjectId _id of the inserted command document.
                'nbErrors': The number of objects that has not been inserted in database due to errors.
            }
        """
        wave = values["Wave"]
        wave_commands_dict = values["Commands"]
        wave_commands = [k for k, v in wave_commands_dict.items() if v == 1]

        self.model.initialize(wave, wave_commands)
        ret, _ = self.model.addInDb()
        # Update this instance.
        return ret, 0  # 0 erros

    def getData(self):
        """Return wave attributes as a dictionnary matching Mongo stored waves
        Returns:
            dict with keys wave, wave_commands, tags and infos
        """
        return self.model.getData()
    def getIntervals(self):
        """Return scope assigned intervals as a list of mongo fetched intervals dict
        Returns:
            list of defect raw mongo data dictionnaries
        """
        return self.model.getIntervals()

    def getChecks(self):
        """Return wave assigned checks as a list of checkInstance
        Returns:
            list of checkInstance objects
        """
        return self.model.getChecks()

    def getAllTools(self):
        """Return all tools being part of this wave as a list of mongo fetched tools dict.
        Differs from getTools as it fetches all tools of the name and not only tools of level wave.
        Returns:
            list of defect raw mongo data dictionnaries
        """
        return self.model.getAllTools()

    def getScopes(self):
        """Return wave assigned scopes as a list of mongo fetched scopes dict
        Returns:
            list of defect raw mongo data dictionnaries
        """
        return self.model.getScopes()

    def isLaunchableNow(self):
        """Returns True if the tool matches criteria to be launched 
        (current time matches one of interval object assigned to this wave)
        Returns:
            bool
        """
        return self.model.isLaunchableNow()

    def getType(self):
        """Returns a string describing the type of object
        Returns:
            "wave" """
        return "wave"





import pollenisatorgui.core.components.utils as utils
import os
from pollenisatorgui.core.components.apiclient import APIClient

def main(apiclient, appli, **kwargs):
	APIClient.setInstance(apiclient)
	if not utils.which_expand_alias("gowitness"):
		return False, "binary 'gowitness' is not in the PATH."
	gowitness_tools = apiclient.find("tools", {"status":"done",  "name":{"$regex": "Gowitness"}})
	export_dir = utils.getExportDir()
	pentest = apiclient.getCurrentPentest().replace("/", "_")
	output_dir = os.path.join(export_dir, "gowitness_server/", pentest)
	os.system("rm -rf  "+str(output_dir))
	merge_str = ""
	for gowitness_tool in gowitness_tools:
		path = apiclient.getResult(str(gowitness_tool["_id"]), output_dir)
		os.system("unzip "+str(path) +" -d "+str(output_dir))
		merge_str += " -i "+path.replace(".zip", ".sqlite3")
	if merge_str == "":
		return False, "No done gotwitness found."
	appli.launch_in_terminal(kwargs.get("default_target",None), "Gowitness server", f"gowitness merge{merge_str} -o {output_dir}/gowitness.sqlite3 & gowitness server -D {output_dir}/gowitness.sqlite3 -P {output_dir}/ & firefox localhost:7171", default_target=kwargs.get("default_target", None))
	return True, f"Opened in firefox"



from pollenisatorgui.core.application.dialogs.ChildDialogAskText import ChildDialogAskText
from pollenisatorgui.core.components import utils


def findDc(apiclient, domain=None, ask_if_fail=True):
    search = {"type":"computer", "infos.is_dc":True}
    if domain is not None:
        search["domain"] = domain
    dc_info = apiclient.find("computers", search, False)
    if dc_info is None:
        if not ask_if_fail:
            return False, "DC not known"
        dialog = ChildDialogAskText(None, "DC not known, give me IP if you know it", multiline=False)
        dialog.app.wait_window(dialog.app)
        dc = dialog.rvalue
    else:
        dc = dc_info.get("ip")
    if dc is None or dc == "":
        return False, "DC not known"
    return True, dc
    

def checkPath(appnames):
    if isinstance(appnames, str):
        appnames = [appnames]
    for appname in appnames:
        path = utils.which_expand_alias(appname)
        if path is not None:
            return True, path
    return False, "App name not found, create an alias or install it. ("+", ".join(appnames)+" were tested)"
    


"""This class pop a defect view form in a subdialog"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.application.dialogs.ChildDialogQuestion import ChildDialogQuestion
from pollenisatorgui.core.application.dialogs.ChildDialogView import ChildDialogView
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.views.defectview import DefectView
from pollenisatorgui.core.controllers.defectcontroller import DefectController
from pollenisatorgui.core.models.defect import Defect
from pollenisatorgui.core.components.utils import get_screen_where_widget

class DummyMainApp:
    def __init__(self, settings):
        self.settings = settings

class ChildDialogDefectView(ChildDialogView):
    """
    Open a child dialog of a tkinter application to answer a question.
    """
    def __init__(self, parent, title, settings, defectModel=None, multi=False, as_template=False, force_insert=False):
        """
        Open a child dialog of a tkinter application to choose autoscan settings.

        Args:
            parent: the tkinter parent view to use for this window construction.
            defectModel : A Defect Model object to load default values. None to have empty fields, default is None.
        """
        super().__init__(parent, title, scrollable=False)
        monitor = get_screen_where_widget(self.app)
        self.app.geometry(f"{monitor.width}x{monitor.height}+{monitor.x}+{monitor.y}")
        self.isInsert = defectModel is None or force_insert
        self.multi = multi
        self.as_template = as_template
        if self.isInsert:
            defectModel = Defect() if defectModel is None else defectModel
            if self.as_template:
                defectModel.isTemplate = True

        self.defect_vw = DefectView(None, self.appFrame, parent,
                                    DefectController(defectModel))
        if self.isInsert:
            if multi:
                self.defect_vw.openMultiInsertWindow(addButtons=False)
                self.completeDialogView(addButtons=True, text_ok="Insert")
            else:
                self.defect_vw.openInsertWindow(addButtons=False)
                self.completeDialogView(addButtons=True, text_ok="Insert")
        else:
            if multi:
                self.defect_vw.openMultiModifyWindow(addButtons=False)
                self.completeDialogView(addButtons=True, text_ok="Update")
            else:
                self.defect_vw.openModifyWindow(addButtons=False)
                self.completeDialogView(addButtons=True, text_ok="Update")
        
    
    def okCallback(self, _event=None):
        """called when pressing the validating button
        Close the window if the form is valid.
        Set rvalue to True and perform the defect update/insert if validated.
        Args:
            _event: Not used but mandatory"""
        msg = None
        if self.isInsert:
            if self.multi:
                res, msg = self.defect_vw.multi_insert()
            else:
                res, msg = self.defect_vw.insert()
        else:
            res, msg = self.defect_vw.update()
        if res:
            self.rvalue = res, msg
            self.app.destroy()
        else:
            tk.messagebox.showerror("Error", msg)
            self.app.destroy()



"""Ip Model. Describes Hosts (not just IP now but domains too)"""

from pollenisatorgui.core.components.datamanager import DataManager
from pollenisatorgui.core.models.element import Element
from bson.objectid import ObjectId
from pollenisatorgui.core.components.apiclient import APIClient
from netaddr import IPNetwork, IPAddress
from netaddr.core import AddrFormatError
from pollenisatorgui.core.models.port import Port
from pollenisatorgui.core.models.tool import Tool
from pollenisatorgui.core.models.defect import Defect
from pollenisatorgui.core.components.utils import isNetworkIp, performLookUp
import re


class Ip(Element):
    """
    Represents an Ip object that defines an Ip or a domain that will be targeted by ip tools.

    Attributes:
        coll_name: collection name in pollenisator database
    """
    coll_name = "ips"

    def __init__(self, valuesFromDb=None):
        """Constructor
        Args:
            valueFromDb: a dict holding values to load into the object. A mongo fetched interval is optimal.
                        possible keys with default values are : _id (None), parent (None), tags([]), infos({}),
                        ip(""), notes(""), in_scopes(None)
        """
        if valuesFromDb is None:
            valuesFromDb = {}
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None),  valuesFromDb.get("infos", {}))
        self.initialize(valuesFromDb.get("ip", ""), valuesFromDb.get("notes", ""),
                        valuesFromDb.get("in_scopes", None), infos=valuesFromDb.get("infos", {}))

    def initialize(self, ip="", notes="", in_scopes=None, infos=None):
        """Set values of ip
        Args:
            ip: the host (ip or domain) to represent
            notes: notes concerning this IP (opt). Default to ""
            in_scopes: a list of scopes that matches this host. If empty this IP will be OOS (Out of Scope). Default to None
            infos: a dictionnary of additional info
        Returns:
            this object
        """
        self.ip = ip
        self.notes = notes
        self.in_scopes = in_scopes if in_scopes is not None else self.getScopesFittingMe()
        self.infos = infos if infos is not None else {}
        return self

    def fitInScope(self, scope):
        """Checks if this IP is the given scope.
        Args:
            scope: a string of perimeter (Network Ip, IP or domain)
        Returns:
            boolean: True if this ip/domain is in given scope. False otherwise.
        """
        if isNetworkIp(scope):
            if Ip.checkIpScope(scope, self.ip):
                return True
        elif Ip.isSubDomain(scope, self.ip):
            return True
        elif self.ip == scope:
            return True
        return False

    def getScopesFittingMe(self):
        """Returns a list of scope objects ids where this IP object fits.
        Returns:
            a list of scopes objects Mongo Ids where this IP/Domain is in scope.
        """
        ret = []
        datamanager = DataManager.getInstance()
        scopes = datamanager.get("scopes",'*')
       
        for scope in scopes:
            if self.fitInScope(scope.scope):
                ret.append(str(scope.getId()))
        return ret


    def delete(self):
        """
        Deletes the Ip represented by this model in database.
        Also deletes the tools associated with this ip
        Also deletes the ports associated with this ip
        Also deletes the defects associated with this ip and its ports
        """
        apiclient = APIClient.getInstance()
        apiclient.delete("ips", ObjectId(self._id))
        

    def addPort(self, values):
        """
        Add a port object to database associated with this Ip.

        Args:
            values: A dictionary crafted by PortView containg all form fields values needed.

        Returns:ret
                '_id': The mongo ObjectId _idret of the inserted port document or None if insertion failed (unicity broken).
        """
        portToInsert = {"ip": self.ip, "port": str(
            values["Port"]), "proto": str(values["Protocole"]), "service": values["Service"], "product": values["Product"]}
        newPort = Port()
        newPort.initialize(
            self.ip, portToInsert["port"], portToInsert["proto"], portToInsert["service"], portToInsert["product"])
        return newPort.addInDb()

    @classmethod
    def checkIpScope(cls, scope, ip):
        """
        Check if the given ip is in the given scope

        Args:
            scope: A network range ip or a domain
            ip: An ipv4 like X.X.X.X

        Returns:
                True if the ip is in the network range or if scope == ip. False otherwise.
        """
        if cls.isIp(scope):
            network_mask = scope.split("/")
            mask = "32"
            if len(network_mask) == 2:
                mask = network_mask[1]
            try:
                res = IPAddress(ip) in IPNetwork(network_mask[0]+"/"+mask)
            except AddrFormatError:
                return False
            except ValueError:
                return False
            return res
        elif scope == ip:
            return True
        return False

    def update(self, pipeline_set=None):
        """Update this object in database.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        if pipeline_set is None:
            apiclient.update("ips", ObjectId(self._id), {"notes": self.notes, "in_scopes": self.in_scopes, "infos": self.infos})
        else:
            apiclient.update("ips", ObjectId(self._id), pipeline_set)

    def addInDb(self):
        """
        Add this IP in database.

        Returns: a tuple with :
                * bool for success
                * mongo ObjectId : already existing object if duplicate, create object id otherwise 
        """
        # Checking unicity
        if self.ip.strip() == "":
            raise ValueError("Ip insertion error: Cannot insert empty ip")
        base = self.getDbKey()
        apiclient = APIClient.getInstance()
        # Add ip as it is unique
        base["notes"] = self.notes
        base["in_scopes"] = self.in_scopes
        base["infos"] = self.infos
        resInsert, idInsert = apiclient.insert("ips", base)
        self._id = idInsert
        # adding the appropriate tools for this port.
        # 1. fetching the wave's commands
        return resInsert, self._id

    

    def __str__(self):
        """
        Get a string representation of an ip.

        Returns:
            Returns the string ipv4 of this ip.
        """
        return self.ip



    def getPorts(self):
        """Returns ip assigned ports as a list of mongo fetched defects dict
        Returns:
            list of port raw mongo data dictionnaries
        """
        datamanager = DataManager.getInstance()
        return datamanager.find("ports", {"ip": self.ip})

    def getDefects(self):
        """Returns ip assigned tools as a list of mongo fetched defects dict
        Returns:
            list of defect raw mongo data dictionnaries
        """
        apiclient = APIClient.getInstance()
        return apiclient.find("defects", {"ip": self.ip, "$or": [{"port": {"$exists": False}}, {"port": None}, {"port": ""}]})

    def getDbKey(self):
        """Return a dict from model to use as unique composed key.
        Returns:
            A dict (1 key :"ip")
        """
        return {"ip": self.ip}

    @classmethod
    def isSubDomain(cls, parentDomain, subDomainTest):
        """Check if the given domain is a subdomain of another given domain
        Args:
            parentDomain: a domain that could be the parent domain of the second arg
            subDomainTest: a domain to be tested as a subdomain of first arg
        Returns:
            bool
        """
        splitted_domain = subDomainTest.split(".")
        # Assuring to check only if there is a domain before the tld (.com, .fr ... )
        topDomainExists = len(splitted_domain) > 2
        if topDomainExists:
            if ".".join(splitted_domain[1:]) == parentDomain:
                return True
        return False

    @classmethod
    def isIp(cls, ip):
        """Checks if the given string is a valid IP
        Args:
            ip: a string that could be representing an ip
        Returns:
            boolean
        """
        return re.search(r"([0-9]{1,3}\.){3}[0-9]{1,3}", ip) is not None

    def getData(self):
        """Returns ip attributes as a dictionnary matching Mongo stored ips
        Returns:
            dict with keys ip, in_scopes, notes, _id, tags and infos
        """
        return {"ip": self.ip, "in_scopes": self.in_scopes, "notes": self.notes, "_id": self.getId(), "infos": self.infos}

    def _getParentId(self):
        """
        Return the mongo ObjectId _id of the first parent of this object.

        Returns:
            Returns the parent's ObjectId _id".
        """
        if self.parent is not None:
            return self.parent
        try:
            if IPAddress(self.ip).is_private():
                return None
        except AddrFormatError:
            return None
        except ValueError:
            return None
        ip_real = performLookUp(self.ip)
        if ip_real is not None:
            datamanager = DataManager.getInstance()
            ip_in_db = datamanager.find("ips", {"ip": ip_real}, False)
            if ip_in_db is None:
                return None
            self.parent = ip_in_db["_id"]
            self.update({"parent": self.parent})
            return ip_in_db["_id"]
        return None



"""Defines a dialog window for choosing 1 between option many thourgh a combobox"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
import pollenisatorgui.core.components.utils as utils

class ChildDialogCombo:
    """
    Open a child dialog of a tkinter application with a combobox.
    """
    def __init__(self, parent, options, displayMsg="Choose a database to open", default=None, **kwargs):
        """
        Open a child dialog of a tkinter application to ask a combobox option.

        Args:
            parent: the tkinter parent view to use for this window construction.
            options: A list of string correspondig to options of the combobox
            displayMsg: The message that will explain to the user what he is choosing.
            default: Choose a default selected option (one of the string in options). default is None
        """
        self.app = CTkToplevel(parent, fg_color=utils.getBackgroundColor())
        self.app.title("Choose option")
        self.app.attributes("-type", "dialog")
        self.app.resizable(False, False)
        self.app.bind("<Escape>", self.onError)
        appFrame = CTkFrame(self.app)
        self.rvalue = None
        self.parent = parent
        if options is None:
            self.onError()
        lbl = CTkLabel(appFrame, text=displayMsg)
        lbl.pack(pady=5)
        kw = {}
        if kwargs.get('width', None) is not None:
            kw["width"] = int(kwargs["width"])
        self.box_template = CTkComboBox(
            appFrame, values=tuple(options), state="readonly",  **kw)
        if default is not None:
            self.box_template.set(default)
        self.box_template.pack(padx=10, pady=5)
        self.box_template.focus_set()
        self.ok_button = CTkButton(appFrame, text="OK", command=self.onOk)
        self.ok_button.bind('<Return>', self.onOk)
        self.ok_button.pack(padx=10, pady=5)
        appFrame.pack(ipadx=10, ipady=5)
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.grab_set()
            self.app.focus_force()
            self.app.lift()
        except tk.TclError:
            pass
        self.box_template.after(50, self.openCombo)

    def openCombo(self):
        self.box_template.event_generate('<Button-1>')

    def onOk(self, event=""):
        """
        Called when the user clicked the validation button. Set the rvalue attributes to the value selected and close the window.
        """
        # send the data to the parent
        self.rvalue = self.box_template.get()
        self.app.destroy()

    def onError(self, _event=None):
        """
        Close the dialog and set rvalue to None
        """
        self.rvalue = None
        self.app.destroy()



"""This class pop a dialog to select checks"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from PIL import ImageTk, Image
from pollenisatorgui.core.application.scrollableframexplateform import ScrollableFrameXPlateform
from pollenisatorgui.core.components.apiclient import APIClient
import pollenisatorgui.core.components.utils as utils
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.models.checkitem import CheckItem
from pollenisatorgui.core.models.command import Command


class ChildDialogSelectChecks:
    """
    Open a child dialog of a tkinter application
    """
    def __init__(self, parent):
        """
        Open a child dialog of a tkinter application to choose autoscan checks.

        Args:
            parent: the tkinter parent view to use for this window construction.
        """
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.title("Choose checks")
        self.app.resizable(True, True)
        self.app.bind("<Escape>", self.cancel)
        self.rvalue = None
        appFrame = ScrollableFrameXPlateform(self.app)
        self.parent = None
        self.initUI(appFrame)
        ok_button = CTkButton(appFrame, text="OK")
        ok_button.pack(side="right", padx=5, pady=10)
        ok_button.bind('<Button-1>', self.okCallback)
        cancel_button = CTkButton(appFrame, text="Cancel", 
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        cancel_button.pack(side="right", padx=5, pady=10)
        cancel_button.bind('<Button-1>', self.cancel)
        appFrame.pack(fill=tk.BOTH, ipady=10, ipadx=10, expand=True)

        self.app.transient(parent)
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.grab_set()
            self.app.focus_force()
            self.app.lift()
        except tk.TclError:
            pass

    def cancel(self, _event=None):
        """called when canceling the window.
        Close the window and set rvalue to False
        Args:
            _event: Not used but mandatory"""
        self.rvalue = None
        self.app.destroy()

    def okCallback(self, _event=None):
        """called when pressing the validating button
        Close the window if the form is valid.
        Set rvalue to True and perform the defect update/insert if validated.
        Args:
            _event: Not used but mandatory"""
        
        self.rvalue = []
        self.app.destroy()

    def initUI(self, parent):
        self.parent = parent
        self.form = FormPanel()
        self.form.addFormLabel("Select checks to queue for autoscan")
        apiclient = APIClient.getInstance()
        check_items = CheckItem.fetchObjects({"check_type":"auto_commands"})
        commands = Command.getList({}, apiclient.getCurrentPentest())
        for check_item in check_items:
            collapsible = self.form.addFormPanel(check_item.title)
            check_commands = [command for command in commands if str(command.original_iid) in check_item.commands]
            for command in check_commands:
                collapsible.addFormCheckbox(command.name, command.name, True)
        self.form.constructView(parent)
            
            


import tkintermd.constants as constants
#import tkintermd.log as log
from customtkinter import *
from pathlib import Path
import tkinter as tk
from tkinter import filedialog, simpledialog
from tkinter import messagebox as mbox
from tkinter.constants import *
import tkinter.ttk as ttk
from tkinterweb import HtmlFrame, Notebook
from tkinterweb.utilities import ScrolledTextBox
from markdown import Markdown
from pygments import lex
from pygments.lexers.markup import MarkdownLexer
from pygments.formatters.html import HtmlFormatter
from pygments.token import Generic
from pygments.lexer import bygroups
from pygments.styles import get_style_by_name, get_all_styles
from pollenisatorgui.core.application.dialogs.ChildDialogAskText import ChildDialogAskText
from pollenisatorgui.core.components import utils

class TkintermdFrame(CTkFrame):
    """A Markdown editor with HTML preview for use in tkinter projects. 
    
    The editor has syntax highlighting supplied by `Pygments` and the HTML 
    preview window is provided by `tkinterweb`.
    
    Import it into your own scripts like so:
    
    ```python
    from tkintermd.frame import TkintermdFrame

    import tkinter as tk
    from tkinter.constants import *

    root = tk.Tk()
    app = TkintermdFrame(root)
    app.pack(fill="both", expand=1)
    app.mainloop()
    ```
    
    """
    def __init__(self, master, **kwargs):
        super().__init__(master,  height=kwargs.get("height", 400))
        #self.logger = log.create_logger()
        self.just_editor = kwargs.get("just_editor", False)
        self.style_change = kwargs.get("style_change", False)
        self.enable_preview = kwargs.get("enable_preview", True)
        self.enable_maximize = kwargs.get("enable_maximize", True)
        self.default = kwargs.get("default_text", "")
        self.binds = kwargs.get("binds", {})
        self.font_size = kwargs.get("font_size", 12)
        # Toolbar.
        if not self.just_editor or self.style_change:
            self.root_toolbar = CTkFrame(self)
            self.open_btn = ttk.Button(self.root_toolbar, text="Open", command=self.open_md_file, tooltip="Open Markdown file")
            self.open_btn.pack(side="left", padx=0, pady=0)
            self.save_as_btn = ttk.Button(self.root_toolbar, text="Save As", command=self.save_as_md_file, tooltip="Save Markdown file as...")
            self.save_as_btn.pack(side="left", padx=0, pady=0)
            self.save_btn = ttk.Button(self.root_toolbar, text="Save", command=self.save_md_file, tooltip="Save Markdown file")
            self.save_btn.pack(side="left", padx=0, pady=0)
            self.export_as_btn = ttk.Button(self.root_toolbar, text="Export HTML", command=self.save_as_html_file, tooltip="Export HTML file as...")
            self.export_as_btn.pack(side="left", padx=0, pady=0)
            # Button to choose pygments style for editor, preview and HTML.
            if self.style_change:
                self.style_opt_btn = tk.Menubutton(self.root_toolbar, text="Editor Style", relief="raised")
                self.style_opt_btn.pack(side="left", padx=0, pady=0)
            self.root_toolbar.pack(side="top", fill="x")

        # Creating the widgets
        self.editor_pw = tk.PanedWindow(self, orient="horizontal", height=kwargs.get("height", 400))
        # Root editor frame
        self.editor_root_frame = tk.Frame(self.editor_pw)
        # Toolbar buttons
        self.editor_toolbar = tk.Frame(self.editor_root_frame)
        self.icons = utils.loadIcons(["maximize.png", "undo_small.png","redo_small.png", "cut_small.png", "copy_small.png", "paste_small.png", "find_small.png","bold_small.png", "italic_small.png", "strikethrough_small.png","image_small.png"])
        if self.enable_maximize:
            self.maximize_btn = ttk.Button(self.editor_toolbar, text="Maximize", image=self.icons["maximize.png"], command=self.maximize, style="icon.TButton", tooltip="Maximize editor")
            self.maximize_btn.pack(side="left", padx=0, pady=0)
        self.undo_btn = ttk.Button(self.editor_toolbar, text="", image=self.icons["undo_small.png"], command=lambda: self.text_area.event_generate("<<Undo>>"), style="icon.TButton", tooltip="Undo last action (Ctrl+z)")
        self.undo_btn.pack(side="left", padx=0, pady=0)
        self.redo_btn = ttk.Button(self.editor_toolbar, text="Redo", image=self.icons["redo_small.png"], command=lambda: self.text_area.event_generate("<<Redo>>"), style="icon.TButton", tooltip="Redo last undo (Ctrl+Shift+z)")
        self.redo_btn.pack(side="left", padx=0, pady=0)
        self.cut_btn = ttk.Button(self.editor_toolbar, text="Cut", image=self.icons["cut_small.png"], command=lambda: self.text_area.event_generate("<<Cut>>"), style="icon.TButton", tooltip="Cut selected text (ctrl+x)")
        self.cut_btn.pack(side="left", padx=0, pady=0)
        self.copy_btn = ttk.Button(self.editor_toolbar, text="Copy", image=self.icons["copy_small.png"], command=lambda: self.text_area.event_generate("<<Copy>>"), style="icon.TButton", tooltip="Copy selected text (ctrl+c)")
        self.copy_btn.pack(side="left", padx=0, pady=0)
        self.paste_btn = ttk.Button(self.editor_toolbar, text="Paste", image=self.icons["paste_small.png"], command=lambda: self.text_area.event_generate("<<Paste>>"), style="icon.TButton", tooltip="Paste selected text (ctrl+v)")
        self.paste_btn.pack(side="left", padx=0, pady=0)
        self.find_btn = ttk.Button(self.editor_toolbar, text="Find", image=self.icons["find_small.png"], command=self.find, style="icon.TButton", tooltip='Find text (ctrl+f)')
        self.find_btn.pack(side="left", padx=0, pady=0)
        self.bold_btn = ttk.Button(self.editor_toolbar, text="", image=self.icons["bold_small.png"], width=16,  command=lambda: self.check_markdown_both_sides(constants.bold_md_syntax, constants.bold_md_ignore, constants.bold_md_special), style="icon.TButton", tooltip='Bold selected text (ctrl+b)')
        self.bold_btn.pack(side="left", padx=0, pady=0)
        self.italic_btn = ttk.Button(self.editor_toolbar, text="", image=self.icons["italic_small.png"], command=lambda: self.check_markdown_both_sides(constants.italic_md_syntax, constants.italic_md_ignore, constants.italic_md_special), style="icon.TButton", tooltip='Italicize selected text (ctrl+i)')
        self.italic_btn.pack(side="left", padx=0, pady=0)
        self.strikethrough_btn = ttk.Button(self.editor_toolbar, text="", image=self.icons["strikethrough_small.png"], command=lambda: self.check_markdown_both_sides(constants.strikethrough_md_syntax, constants.strikethrough_md_ignore, md_special=None, strikethrough=True), style="icon.TButton", tooltip='Strikethrough selected text (ctrl+t)')
        self.strikethrough_btn.pack(side="left", padx=0, pady=0)
        # self.heading_btn = ttk.Button(self.editor_toolbar, text="Heading")
        # self.heading_btn.pack(side="left", padx=0, pady=0)
        # self.unordered_list_btn = ttk.Button(self.editor_toolbar, text="Unordered List")
        # self.unordered_list_btn.pack(side="left", padx=0, pady=0)
        # self.ordered_list_btn = ttk.Button(self.editor_toolbar, text="Ordered List")
        # self.ordered_list_btn.pack(side="left", padx=0, pady=0)
        # self.checklist_btn = ttk.Button(self.editor_toolbar, text="Checklist")
        # self.checklist_btn.pack(side="left", padx=0, pady=0)
        # self.blockquote_btn = ttk.Button(self.editor_toolbar, text="Blockquote")
        # self.blockquote_btn.pack(side="left", padx=0, pady=0)
        #self.codeblock_btn = ttk.Button(self.editor_toolbar, text="Codeblock", command=lambda: self.check_markdown_both_sides(("`","``","```"), constants.bold_md_ignore), style="icon.TButton")
        #self.codeblock_btn.pack(side="left", padx=0, pady=0)
        # self.table_btn = ttk.Button(self.editor_toolbar, text="Table")
        # self.table_btn.pack(side="left", padx=0, pady=0)
        # self.link_btn = ttk.Button(self.editor_toolbar, text="Link")
        # self.link_btn.pack(side="left", padx=0, pady=0)
        self.image_btn =  ttk.Button(self.editor_toolbar, text="", image=self.icons["image_small.png"], command=lambda: self.add_image(), style="icon.TButton", tooltip='Add image (you can also drag&drop inside the editor)')
        self.image_btn.pack(side="left", padx=0, pady=0)
        self.font_size_label = ttk.Label(self.editor_toolbar, text="Font size:")
        self.font_size_label.pack(side="left", padx=0, pady=0)
        self.font_size_combo = ttk.Combobox(self.editor_toolbar, values=list(range(6,64,2)), width=3)
        self.font_size_combo.set(self.font_size)
        self.font_size_combo.pack(side="left", padx=0, pady=0)
        self.font_size_combo.bind("<<ComboboxSelected>>", self.on_font_size_change)
        self.editor_toolbar.pack(side="top", fill="x")
        # Editor frame with scrollbar and text area.
        
        self.editor_frame = ScrolledTextBox(self.editor_root_frame, undo=True, wrap=tk.WORD, font=("Times New Roman", self.font_size))
        self.text_area = self.editor_frame.tbox
        self.text_area.register_drop_target("*")
        if self.binds.get("<<Drop:File>>", None) is not None:
            self.text_area.bind('<<Drop:File>>', self.binds.get("<<Drop:File>>"))
        else:
            self.text_area.bind('<<Drop:File>>', self.dropFile)

        self.editor_frame.pack(fill="both", expand=1)
        # Tabs for the preview and export options.
        if self.enable_preview:
            self.preview_tabs = Notebook(self.editor_pw)
            # HTML rendered preview.
            self.preview_document = HtmlFrame(self.preview_tabs, messages_enabled=False)
        # Root frame for export options.
        self.export_options_root_frame = tk.Frame(self.preview_tabs)
        # Export options frame.
        self.export_options_frame = tk.Frame(self.export_options_root_frame)
        # self.export_options_placeholder = tk.Label(self.export_options_frame, text="Placeholder", justify="center")
        # self.export_options_placeholder.pack(fill="both", expand=1)
        self.export_options_row_a = tk.Frame(self.export_options_frame)
        self.template_label = tk.Label(self.export_options_row_a, text="Choose template:\t")
        self.template_label.pack(side="left")
        self.template_Combobox_value = tk.StringVar()
        self.template_Combobox = ttk.Combobox(self.export_options_row_a, textvariable=self.template_Combobox_value, values=constants.template_list)
        self.template_Combobox.current(0)
        self.template_Combobox.pack(side="left")
        self.export_options_edit_btn = ttk.Button(self.export_options_row_a, text="Edit Before Export", command=self.enable_edit)
        self.export_options_edit_btn.pack(side="left", padx=0, pady=0)
        self.export_options_export_btn = ttk.Button(self.export_options_row_a, text="Export HTML", command=self.save_as_html_file)
        self.export_options_export_btn.pack(side="left", padx=0, pady=0)
        self.export_options_row_a.pack(fill="both")
        self.export_options_frame.pack(fill="both")
        # HTML code preview/edit before export with scrollbar and text area.
        self.export_options_text_area_frame = ScrolledTextBox(self.export_options_root_frame)
        self.export_options_text_area = self.export_options_text_area_frame.tbox
        self.export_options_text_area.configure(state="disabled")
        self.export_options_text_area_frame.pack(fill="both", expand=1)
        # Add the rendered preview and export options to the Notebook tabs.
        self.preview_tabs.add(self.preview_document, text="Preview Document")
        self.preview_tabs.add(self.export_options_root_frame, text="Export Options")
        # Add the editor and preview/export areas to the paned window.
        self.editor_pw.add(self.editor_root_frame)
        self.editor_pw.add(self.preview_tabs)
        self.editor_pw.pack(side="top", fill="both", expand=1)
        self.css = ""
        if self.style_change:
            # Load the self.style_opt_btn menu, this needs to be after the editor.
            self.style_menu = tk.Menu(self.style_opt_btn, tearoff=False)
            for style_name in get_all_styles():
                # dynamcially get names of styles exported by pygments
                try:
                    # test them for compatability
                    self.load_style(style_name)
                    
                except Exception as E:
                    #self.logger.exception(f"WARNING: style {style_name} failed ({E}), removing from style menu.")
                    continue # don't add them to the menu
                # add the rest to the menu
                self.style_menu.add_command(
                    label=style_name,
                    command=(lambda sn:lambda: self.load_style(sn))(style_name)
                    # unfortunately lambdas inside loops need to be stacked 2 deep to get closure variables instead of cell variables
                    )
            self.style_opt_btn["menu"] = self.style_menu

        # Set Pygments syntax highlighting style.
        self.lexer = Lexer()
        self.syntax_highlighting_tags = self.load_style("stata-dark")
        # self.syntax_highlighting_tags = self.load_style("material")
        # Default markdown string.
        default_text = self.default
        default_text = default_text.replace("\t", "    ")
        default_text = default_text.replace("\r", "")
        self.text_area.insert(0.0, default_text)
        self.template_top = constants.default_template_top
        self.template_middle = constants.default_template_middle
        self.template_bottom = constants.default_template_bottom
        # Applies markdown formatting to default file.
        self.check_markdown_highlighting(start="1.0", end=END)
        self.text_area.focus_set()

        # Create right click menu layout for the editor.
        self.right_click = tk.Menu(self.text_area, tearoff=False)
        self.right_click.add_command(label="Copy", command=lambda: self.focus_get().event_generate("<<Copy>>"), accelerator="Ctrl+C")
        self.right_click.add_command(label="Cut", command=lambda: self.focus_get().event_generate("<<Cut>>"), accelerator="Ctrl+X")
        self.right_click.add_command(label="Paste", command=lambda: self.focus_get().event_generate("<<Paste>>"), accelerator="Ctrl+V")
        self.right_click.add_separator()
        self.right_click.add_command(label="Undo", command=lambda: self.focus_get().event_generate("<<Undo>>"), accelerator="Ctrl+Z")
        self.right_click.add_command(label="Redo", command=lambda: self.focus_get().event_generate("<<Redo>>"), accelerator="Ctrl+Y")
        self.right_click.add_separator()
        self.right_click.add_command(label="Find", command=self.find, accelerator="Ctrl+F")
        self.right_click.add_command(label="Select All", command=self.select_all, accelerator="Ctrl+A")

        # Bind mouse/key events to functions.
        self.template_Combobox.bind("<<ttk.ComboboxSelected>>", self.change_template)
        self.text_area.bind("<<Modified>>", self.on_input_change)
        self.text_area.edit_modified(0)#resets the text widget to generate another event when another change occours
        
        self.text_area.bind_all("<Control-f>", self.find)
        self.text_area.bind_all("<Control-a>", self.select_all)
        self.text_area.bind("<Button-3>", self.popup)

        # # This links the scrollbars but is currently causing issues. 
        # Changing the settings to make the scrolling work
        # self.preview_document.html['yscrollcommand'] = self.on_mousewheel

    def on_font_size_change(self, event=None):
        """Changes the font size of the text area.
        
        - Queries and changes the font size of the text area.
        """
        font_size = int(self.font_size_combo.get())
        self.font_size = font_size
        self.load_style(self.stylename)
        scale_factor = font_size/12.0
        self.preview_document.set_fontscale(scale_factor)

    def dropFile(self, event):
        # This function is called, when stuff is dropped into a widget
        local_paths = utils.drop_file_event_parser(event)
        for local_path in local_paths:
            self.text_area.insert(tk.INSERT, f"![{local_path}]({local_path})")

    def popup(self, event):
        """Right-click popup at mouse location within the text area only.
        
        Provides the following options:
        
        - Cut.
        - Copy.
        - Paste.
        - Undo.
        - Redo.
        - Find.
        - Select All.
        """
        self.right_click.tk_popup(event.x_root, event.y_root)

    # def on_scrollbar(self, *args):
    #     """Scrolls the text area scrollbar when clicked/dragged with a mouse.
        
    #     - Queries and changes the vertical position of the text area view.
    #     """
    #     self.text_area.yview(*args)
    #     # # This links the scrollbars but is currently causing issues.
    #     # self.preview_document.html.yview(*args)

    # def on_mousewheel(self, *args):
    #     """Moves the scrollbar and scrolls the text area on mousewheel event.
        
    #     - Sets the fractional values of the slider position (upper and lower 
    #         ends as value between 0 and 1).
    #     - Calls `on_scrollbar` function.
    #     """
    #     self.scrollbar.set(*args)
    #     # # This links the scrollbars but is currently causing issues.
    #     # self.preview_document.vsb.set(*args)
    #     self.on_scrollbar('moveto', args[0])

    def select_all(self, *args):
        """Select all text within the editor window.
        
        - Add tag TAGNAME to all characters between INDEX1 and INDEX2.
        - Set mark MARKNAME before the character at index.
        - Scroll so that the character at INDEX is visible.
        """
        self.text_area.tag_add(SEL, "1.0", END)
        self.text_area.mark_set(0.0, END)
        self.text_area.see(INSERT)

    def maximize(self, *args, **kwargs):
        """Maximize the editor window.
        
        - Maximizes the editor window.
        """
        dialog = ChildDialogAskText(self, "Markdown Editor", self.text_area.get("1.0", END), multiline=True, markdown=True, fullscreen=True, save_on_close=True)
        self.wait_window(dialog.app)
        text = dialog.rvalue
        if text is not None:
            self.text_area.delete("1.0", END)
            self.text_area.insert("1.0", text)

    def find(self, *args):
        """Simple search dialog for within the editor window.
        
        - Get the current text area content.
        - Displays a simple dialog with a field to enter a search string into 
            and two buttons.
        - If a string is provided then search for it within the text area.
        - Add tag TAGNAME to all characters between INDEX1 and INDEX2.
        - Highlight any found strings within the text editor.
        """
        self.text_area.tag_remove('found', '1.0', END)
        target = simpledialog.askstring('Find', 'Search String:')

        if target:
            idx = '1.0'
            while 1:
                idx = self.text_area.search(target, idx, nocase=1, stopindex=END)
                if not idx: break
                lastidx = '%s+%dc' % (idx, len(target))
                self.text_area.tag_add('found', idx, lastidx)
                idx = lastidx
            self.text_area.tag_config('found', foreground='white', background='blue')

    def open_md_file(self):
        """Open a file and clear/insert the text into the text_area.
        
        Opens a native OS dialog and expects markdown formatted files. Shows an 
        error message if it fails.

        - Display a native OS dialog and request a filename to open.
        - If a filename is provided then `try` to open it in "read" mode.
        - Replace the text area content.
        - Set the `constants.cur_file` value to the filename that was opened.
        - If any of the above fails then display an error message.
        """
        open_filename_md = filedialog.askopenfilename(filetypes=(("Markdown File", "*.md , *.mdown , *.markdown"), ("Text File", "*.txt"), ("All Files", "*.*")))
        if open_filename_md:
            try:
                with open(open_filename_md, "r") as stream:
                    open_filename_contents = stream.read()
                self.text_area.delete(1.0, END)
                self.text_area.insert(END, open_filename_contents)
                self.check_markdown_highlighting(start="1.0", end=END)
                constants.cur_file = Path(open_filename_md)
            except:
                mbox.showerror(title="Error", message=f"Error Opening Selected File\n\nThe file you selected: {open_filename_md} can not be opened!")
    
    def save_as_md_file(self):
        """Saves the file with the given filename.
        
        Opens a native OS dialog for saving the file with a name and markdown 
        extension. Shows an error message if it fails.

        - Display a native OS dialog and request a filename to save.
        - If a filename is provided then `try` to open it in "write" mode.
        - Set the `constants.cur_file` value to the filename that was opened.
        - If any of the above fails then display an error message.
        """
        self.file_data = self.text_area.get("1.0" , END)
        self.save_filename_md = filedialog.asksaveasfilename(filetypes = (("Markdown File", "*.md"), ("Text File", "*.txt")) , title="Save Markdown File")
        if self.save_filename_md:
            try:
                with open(self.save_filename_md, "w") as stream:
                    stream.write(self.file_data)
                    constants.cur_file = Path(self.save_filename_md)
            except:
                mbox.showerror(title="Error", message=f"Error Saving File\n\nThe file: {self.save_filename_md} can not be saved!")

    def save_md_file(self):
        """Quick saves the file with its current name.

        - Get the current text area content.
        - `try` to save the file with the `constants.cur_file` variable.
        - If it fails because no name exists it calls the "save_as_md_file" 
            function.
        """
        self.file_data = self.text_area.get("1.0" , END)
        try:
            with open(constants.cur_file, "w") as stream:
                stream.write(self.file_data)
        except:
            self.save_as_md_file()

    def save_as_html_file(self):
        """Exports the current contents of the HTML preview pane to the given filename.
        
        Opens a native OS dialog for saving the file with a html
        extension. Shows an error message if it fails.

        - Display a native OS dialog and request a filename to save.
        - If a filename is provided then `try` to open it in "write" mode.
        - If any of the above fails then display an error message.
        """
        html_file_data = self.export_options_text_area.get("1.0" , END)
        self.html_save_filename = filedialog.asksaveasfilename(filetypes = (("HTML file", ("*.html", "*.htm")),) , title="Save HTML File")
        if self.html_save_filename:
            try:
                with open(self.html_save_filename, "w") as stream:
                    stream.write(html_file_data)
                    #constants.cur_file = Path(self.html_save_filename)
            except:
                mbox.showerror(title="Error", message=f"Error Saving File\n\nThe file: {self.html_save_filename} can not be saved!")

    def on_input_change(self, event):
        return self.do_on_input_change(event)    
    def do_on_input_change(self, event):
        """Converts the text area input into html output for the HTML preview.
        
        When the user types:
        
        - Get the current text area contents.
        - Convert the markdown formatted string to HTML.
        - Merge the converted markdown with the currently selected template.
        - Load the merged HTML into the document preview.
        - Update the HTML content/states within the export options edit area.
        - Check the markdown and apply formatting to the text area.
        - Reset the modified flag.
        """
        md2html = Markdown(extensions=constants.extensions, extension_configs=constants.extension_configs)
        markdownText = self.text_area.get("1.0", END)
        html = md2html.convert(markdownText)
        final = f"{self.template_top}\n{self.css}\n{self.template_middle}\n{html}\n{self.template_bottom}"
        self.export_options_text_area.configure(state="normal")
        self.export_options_text_area.delete("1.0" , END)
        self.export_options_text_area.insert(END, final)
        self.export_options_text_area.configure(state="disabled")
        self.export_options_edit_btn.configure(state="normal")
        try:
            self.preview_document.load_html(final)
        except RuntimeError as e:
            pass #BUG, understand why main thread is not in main loop
        # self.preview_document.add_css(self.css)
        self.check_markdown_highlighting(start="1.0", end=END)
        self.text_area.edit_modified(0) # resets the text widget to generate another event when another change occours

    def load_style(self, stylename):
        self.stylename = stylename
        return self.do_load_style(stylename)

    def do_load_style(self, stylename):
        """Load Pygments style for syntax highlighting within the editor.
        
        - Load and configure the text area and `tags` with the Pygments styles 
            and custom `Lexer` tags.
        - Create the CSS styling to be merged with the HTML template
        - Generate a `<<Modified>>` event to update the styles when a new style 
            is chosen.
        """
        self.style = get_style_by_name(stylename)
        self.syntax_highlighting_tags = []
        for token, opts in self.style.list_styles():
            kwargs = {}
            fg = opts['color']
            bg = opts['bgcolor']
            if fg:
                kwargs['foreground'] = '#' + fg
            if bg:
                kwargs['background'] = '#' + bg
            font = ('Monospace', self.font_size) + tuple(key for key in ('bold', 'italic') if opts[key])
            kwargs['font'] = font
            kwargs['underline'] = opts['underline']
            self.text_area.tag_configure(str(token), **kwargs)
            self.export_options_text_area.tag_configure(str(token), **kwargs)
            self.syntax_highlighting_tags.append(str(token))
        # print(self.style.background_color or 'white', self.text_area.tag_cget("Token.Text", "foreground") or 'black', stylename)
        self.text_area.configure(bg=self.style.background_color or 'white',
                        fg=self.text_area.tag_cget("Token.Text", "foreground") or 'black',
                        selectbackground=self.style.highlight_color,
                        insertbackground=self.text_area.tag_cget("Token.Text", "foreground") or 'black',
                        )
        self.text_area.tag_configure(str(Generic.StrongEmph), font=('Monospace', self.font_size, 'bold', 'italic'))
        self.export_options_text_area.configure(bg=self.style.background_color or 'white',
                        fg=self.export_options_text_area.tag_cget("Token.Text", "foreground") or 'black',
                        selectbackground=self.style.highlight_color,
                        insertbackground=self.text_area.tag_cget("Token.Text", "foreground") or 'black',
                        )
        # self.export_options_text_area.tag_configure(str(Generic.StrongEmph), font=('Monospace', self.font_size, 'bold', 'italic'))
        self.syntax_highlighting_tags.append(str(Generic.StrongEmph))
        self.formatter = HtmlFormatter()
        self.pygments = self.formatter.get_style_defs(".highlight")
        # Previous version.
        self.css = 'body {background-color: %s; color: %s }\nbody .highlight{ background-color: %s; }\n%s' % (
            self.style.background_color,
            self.text_area.tag_cget("Token.Text", "foreground"),
            self.style.background_color,
            self.pygments
            )#used string%interpolation here because f'string' interpolation is too annoying with embeded { and }
        # self.preview_document.add_css(self.css)
        self.text_area.event_generate("<<Modified>>")
        from pollenisatorgui.core.components.settings import Settings
        settings = Settings()
        editor_key_style = "editor_dark_style" if settings.is_dark_mode() else "editor_light_style"
        settings.local_settings[editor_key_style] = stylename
        settings.saveLocalSettings()
        return self.syntax_highlighting_tags    

    def check_markdown_highlighting(self, start='insert linestart', end='insert lineend'):
        self.do_check_markdown_highlighting(start, end)

    def do_check_markdown_highlighting(self, start='insert linestart', end='insert lineend'):
        """Formats editor content using the Pygments style."""
        self.data = self.text_area.get(start, end)
        while self.data and self.data[0] == '\n':
            start = self.text_area.index('%s+1c' % start)
            self.data = self.data[1:]
        self.text_area.mark_set('range_start', start)
        # clear tags
        for t in self.syntax_highlighting_tags:
            self.text_area.tag_remove(t, start, "range_start +%ic" % len(self.data))
        # parse text
        for token, content in lex(self.data, self.lexer):
            self.text_area.mark_set("range_end", "range_start + %ic" % len(content))
            for t in token.split():
                self.text_area.tag_add(str(t), "range_start", "range_end")
            self.text_area.mark_set("range_start", "range_end")

    def apply_markdown_both_sides(self, selection, md_syntax):
        """Apply markdown to both sides of a selection.

        Args:
            selection (str): Text selection from the editor.
            md_syntax (tuple): Tuple of markdown strings to apply.
        """
        self.md_syntax = md_syntax
        self.cur_selection = selection
        self.insert_md = f"{self.md_syntax[0]}{self.cur_selection}{self.md_syntax[0]}"
        self.text_area.delete(index1=SEL_FIRST, index2=SEL_LAST)
        self.text_area.insert(INSERT, self.insert_md)
        return

    def remove_markdown_both_sides(self, selection, md_syntax):
        """Remove markdown from both sides of a selection.

        Args:
            selection (str): Text selection from the editor.
            md_syntax (tuple): Tuple of markdown strings to remove.
        """
        self.md_syntax = md_syntax
        self.cur_selection = selection
        self.remove_md = str(self.cur_selection).strip(self.md_syntax[0]).strip(self.md_syntax[1])
        self.text_area.delete(index1=SEL_FIRST, index2=SEL_LAST)
        self.text_area.insert(INSERT, self.remove_md)
        return

    def add_image(self):
        """add image to the editor."""
        image_path = filedialog.askopenfilename(title = "Select file",filetypes = (("jpeg files","*.jpg"),("png files","*.png"),("all files","*.*")))
        if image_path:
            self.text_area.insert(INSERT, f"![{os.path.basename(image_path)}]({image_path})")
        return

    def check_markdown_both_sides(self, md_syntax, md_ignore, md_special, strikethrough=None):
        """Check markdown formatting to be applied to both sides of a selection. 

        This will ignore items in the md_ignore variable and then deal with 
        special syntax individually before applying or removing the markdown 
        formatting.
        
        - If string starts with anything in md_ignore do nothing and return from
        the function.
        - If strikethrough is set to `True` then apply or remove the markdown.
        - If the formatting requires special items which can't go in md_ignore
        because they cause issues with markdown being applied incorrectly do 
        nothing and return from the function.
        - Apply or remove the markdown once we reach the end.

        Args:
            selection (str): Text selection from the editor.
            md_syntax (tuple): Tuple of markdown strings to remove.
            md_ignore (tuple): Tuple of markdown strings to ignore.
            md_special (tuple): Tuple of special markdown strings to ignore that 
                cause unexpected issues when included in md_ignore.
            strikethrough (bool): Set to True for strikethrough. Default is 
                `None`.
        """
        self.md_syntax = md_syntax
        self.md_ignore = md_ignore
        self.md_special = md_special
        self.cur_selection = self.text_area.selection_get()
        if str(self.cur_selection).startswith(self.md_ignore) or str(self.cur_selection).endswith(self.md_ignore):
            return
        elif strikethrough == True:
            if str(self.cur_selection).startswith(self.md_syntax) and str(self.cur_selection).endswith(self.md_syntax):
                self.remove_markdown_both_sides(self.cur_selection, self.md_syntax)
                return
            else:
                self.apply_markdown_both_sides(self.cur_selection, self.md_syntax)
                return
        elif str(self.cur_selection).startswith(self.md_special) and str(self.cur_selection).endswith(self.md_special) and not str(self.cur_selection).startswith(self.md_syntax) and not str(self.cur_selection).startswith(self.md_syntax):
            return 
        elif str(self.cur_selection).startswith(self.md_syntax) and str(self.cur_selection).endswith(self.md_syntax):
            self.remove_markdown_both_sides(self.cur_selection, self.md_syntax)
        else:
            self.apply_markdown_both_sides(self.cur_selection, self.md_syntax)

    def change_template(self, template_name):
        """Change the currently selected template.
        
        Get the selected template name from the `StringVar` for the `ttk.Combobox` 
        and compare it with the templates dictionary. If the name matches the 
        key then set the relevant template values and update all the previews.
        """
        template_name = self.template_Combobox_value.get()
        for key, val in constants.template_dict.items():
            if template_name == key:
                self.template_top = val[0]
                self.template_middle = val[1]
                self.template_bottom = val[2]
        self.text_area.event_generate("<<Modified>>")
    
    def enable_edit(self):
        """Enable editing of HTML before export.
        
        Displays a warning to the user and enables HTML editing prior to export.
        """
        mbox.showwarning(title="Warning", message=constants.edit_warning)
        self.export_options_text_area.configure(state="normal")
        self.export_options_edit_btn.configure(state="disabled")

class Lexer(MarkdownLexer):
    """Extend MarkdownLexer to add markup for bold-italic. 
    
    This needs extending further before being complete.
    """
    tokens = {key: val.copy() for key, val in MarkdownLexer.tokens.items()}
    # # bold-italic fenced by '***'
    tokens['inline'].insert(2, (r'(\*\*\*[^* \n][^*\n]*\*\*\*)',
                                bygroups(Generic.StrongEmph)))
    # # bold-italic fenced by '___'
    tokens['inline'].insert(2, (r'(\_\_\_[^_ \n][^_\n]*\_\_\_)',
                                bygroups(Generic.StrongEmph)))



from pollenisatorgui.core.plugins.plugin import Plugin

class Default(Plugin):
    """Attributes:
        autoDefect: a boolean indication that this plugin should be used to autodetect file.
    """
    autoDetect = False  # Override default True value

def getActions():
    return ""


"""Command Model."""
from pollenisatorgui.core.models.element import Element
from pollenisatorgui.core.components.apiclient import APIClient


class Command(Element):
    """Represents a command object to be run on designated scopes/ips/ports.

    Attributes:
        coll_name: collection name in pollenisator or pentest database
    """

    coll_name = "commands"

    def __init__(self, valuesFromDb=None):
        """Constructor
        Args:
            valueFromDb: a dict holding values to load into the object. A mongo fetched command is optimal.
                        possible keys with default values are : _id (None), parent (None), tags([]), infos({}), name(""), 
                         text(""), lvl("port:onServiceUpdate"), ports(""), safe(True), types([]), indb="pollenisator", timeout="300", owners=[]
        """
        if valuesFromDb is None:
            valuesFromDb = dict()
        super().__init__(valuesFromDb.get("_id", None), valuesFromDb.get("parent", None), valuesFromDb.get("infos", {}))
        self.initialize(valuesFromDb.get("name", ""), valuesFromDb.get("bin_path", ""), valuesFromDb.get("plugin", ""), 
                        valuesFromDb.get("text", ""), 
                        valuesFromDb.get("indb", "pollenisator"),  valuesFromDb.get("original_iid"), valuesFromDb.get("timeout", 300), 
                        valuesFromDb.get("owners", []), valuesFromDb.get("infos", {}))

    def initialize(self, name, bin_path, plugin="Default", text="", indb=False, original_iid="None", timeout=300, owners=[], infos=None):
        """Set values of command
        Args:
            name: the command name
            bin_path: local command, binary path or command line
            plugin: plugin that goes with this command
            text: the command line options. Default is "".
            indb: db name : global (pollenisator database) or  local pentest database
            timeout: a timeout to kill stuck tools and retry them later. Default is 300 (in seconds)
            owners: the owners of the command
            infos: a dictionnary with key values as additional information. Default to None
        Returns:
            this object
        """
        self.name = name
        self.bin_path = bin_path
        self.plugin = plugin
        self.text = text
        self.original_iid = original_iid
        self.infos = infos if infos is not None else {}
        self.indb = indb
        self.timeout = timeout
        self.owners = owners
        return self

    def delete(self):
        """
        Delete the command represented by this model in database.
        Also delete it from every waves's wave_commands
        Also delete every tools refering to this command.
        """
        ret = self._id
        apiclient = APIClient.getInstance()
        if self.indb == "pollenisator":
            apiclient.deleteCommand(ret)
        else:
            apiclient.delete("commands", ret)
        

    def addInDb(self):
        """Add this command to pollenisator database
        Returns: a tuple with :
                * bool for success
                * mongo ObjectId : already existing object if duplicate, create object id otherwise 
        """
        apiclient = APIClient.getInstance()
        res, id = apiclient.insert("commands", {"name": self.name, "bin_path":self.bin_path, "plugin":self.plugin, 
                                                                            "text": self.text,
                                                                           "indb": self.indb, "timeout": int(self.timeout), "owners": self.owners})
        if not res:
            return False, id
        self._id = id
        return True, id
        

    def update(self, pipeline_set=None):
        """Update this object in database.
        Args:
            pipeline_set: (Opt.) A dictionnary with custom values. If None (default) use model attributes.
        """
        apiclient = APIClient.getInstance()
        
        if pipeline_set is None:
            apiclient.update("commands", self._id, {"timeout": int(self.timeout),
                         "text": self.text, "indb":self.indb, "plugin":self.plugin, "bin_path":self.bin_path})
           
        else:
            apiclient.update("commands", self._id, pipeline_set)
            

    @classmethod
    def getList(cls, pipeline=None, targetdb="pollenisator"):
        """
        Get all command's name registered on database
        Args:
            pipeline: default to None. Condition for mongo search.
        Returns:
            Returns the list of commands name found inside the database. List may be empty.
        """
        if pipeline is None:
            pipeline = {}
        return [command for command in cls.fetchObjects(pipeline, targetdb)]

    @classmethod
    def fetchObject(cls, pipeline, targetdb="pollenisator"):
        """Fetch one command from database and return the Command object 
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a Command or None if nothing matches the pipeline.
        """
        apiclient = APIClient.getInstance()
        if targetdb == "pollenisator":
            res = apiclient.findCommand(pipeline)
            d = res[0] if res else None
        else:
            d = apiclient.findInDb(targetdb, "commands", pipeline, False)
        if d is None:
            return None
        return Command(d)

    @classmethod
    def fetchObjects(cls, pipeline, targetdb="pollenisator"):
        """Fetch many commands from database and return a Cursor to iterate over Command objects
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a cursor to iterate on Command objects
        """
        apiclient = APIClient.getInstance()
        if targetdb == "pollenisator":
            ds = apiclient.findCommand(pipeline)
        else:
            ds = apiclient.findInDb(targetdb, "commands", pipeline, True)
        if ds is None:
            return None
        for d in ds:
            yield Command(d)
    
    @classmethod
    def getMyCommands(cls, user=None):
        """Fetch many commands from database and return a Cursor to iterate over Command objects
        Args:
            pipeline: a Mongo search pipeline (dict)
        Returns:
            Returns a cursor to iterate on Command objects
        """
        apiclient = APIClient.getInstance()
        if user is None:
            user = apiclient.getUser()
        ds = apiclient.findCommand({"owners":user})
        if ds is None:
            return None
        for d in ds:
            yield Command(d)    

    def addToMyCommands(self):
        apiclient = APIClient.getInstance()
        apiclient.addCommandToMyCommands(self.getId())

    def isMine(self):
        user = APIClient.getInstance().getUser()
        return user in self.owners
        
    def __str__(self):
        """
        Get a string representation of a command.

        Returns:
            Returns the command's name string.
        """
        return self.name

    def getData(self):
        """Return command attributes as a dictionnary matching Mongo stored commands
        Returns:
            dict with keys name, lvl, safe, text, ports, priority, max_thread, priority, types, _id, tags and infos
        """
        return {"name": self.name, "bin_path":self.bin_path, "plugin":self.plugin,  "text": self.text,
                "timeout": self.timeout,
                "indb":self.indb, "owners": self.owners, "_id": self.getId(), "infos": self.infos}


    def getDbKey(self):
        """Return a dict from model to use as unique composed key.
        Returns:
            A dict (1 key :"name")
        """
        return {"name": self.name}




"""Show a progess bar for the user.
"""
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *


class ChildDialogProgress:
    """
    Open a child dialog of a tkinter application to inform the user about a ongoing process.
    """

    def __init__(self, parent, title, msg, speed=1, progress_mode="determinate", show_logs=False):
        """
        Open a child dialog of a tkinter application to display a progress bar.

        Args:
            parent: the tkinter parent view to use for this window construction.
            title: Title for the new window
            msg: Message to display on the window to inform about a progession.
            progress_mode: mode of progression. Either "determinate" or "inderterminate". Default to the second.
                           indeterminate: bouncing progress bar.
                           determinate: Show progression of a value against a max value.
        """
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.transient(parent)
        self.app.resizable(False, False)
        self.app.title(title)
        appFrame = CTkFrame(self.app)
        self.rvalue = None
        self.parent = parent
        self.lbl = CTkLabel(appFrame, text=msg)
        self.lbl.pack(side=tk.TOP, padx=10, pady=10, fill=tk.X)
        self.mode = progress_mode
        self.show_logs = show_logs
        if self.show_logs:
            self.text_log = CTkTextbox(
                appFrame, wrap="word")
            self.text_log.pack(side=tk.BOTTOM, padx=10,pady=10,fill=tk.X)
        if progress_mode == "determinate":
            speed = (100/speed)*50/100
        self.progressbar = CTkProgressBar(appFrame, orientation="horizontal",
                                           indeterminate_speed=speed, determinate_speed=speed, mode=progress_mode)
        self.progressbar.set(0)
        self.progressbar.pack(side=tk.BOTTOM, padx=10, pady=10, fill=tk.X)
        appFrame.pack(fill=tk.BOTH)
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.focus_force()
            #self.app.grab_set()
            self.app.lift()
        except tk.TclError:
            pass

    def show(self, maximum=None, startValue=0):
        """Start displaying the progressbar.
        Args:
            - maximum: only for determinate mode. Set the goal value. Default to None.
            - startValue: only for determinate mode. Set the starting value. Default to None.
        """
        self.progressbar.start()
        self.app.update()


    def update(self, value=None, msg=None, log=None):
        """Update the progressbar and show progression value.
        Call this regularly if on inderminate mode.
        Args:
            - value: The new value for the progressbar. Default to None.
        """
        try:
            if self.mode == "indeterminate":
                try:
                    self.progressbar.step()
                except tk.TclError:
                    return
            elif self.mode == "determinate":
                self.progressbar.step()
            if msg is not None:
                self.lbl.configure(text=str(msg))
            if self.show_logs and log is not None:
                self.text_log.insert(tk.END, log)
                self.text_log.see(tk.END)
            self.app.update()
        except tk.TclError as e:
            #Probably exited
            raise e

    def destroy(self):
        """
        Close the window and stop the progressbar.
        """
        # send the data to the parent
        self.progressbar.stop()
        self.app.destroy()



"""Defines a sub-swindow window for connecting to the server"""

import tkinter as tk
import tkinter.ttk as ttk
import re
from customtkinter import *
from pollenisatorgui.core.application.pollenisatorentry import PopoEntry
from PIL import ImageTk, Image
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.components.utils import loadClientConfig, saveClientConfig, getValidMarkIconPath, getBadMarkIconPath, getWaitingMarkIconPath
from pollenisatorgui.core.application.CollapsibleFrame import CollapsibleFrame

class ChildDialogConnect:
    """
    Open a child dialog of a tkinter application to ask server and login infos
    """
    cvalid_icon = None
    cbad_icon = None
    cwaiting_icon = None

    def validIcon(self):
        """Returns a icon indicating a valid state.
        Returns:
            ImageTk PhotoImage"""
        if self.__class__.cvalid_icon is None:
            self.__class__.cvalid_icon = CTkImage(
                Image.open(getValidMarkIconPath()))
        return self.__class__.cvalid_icon

    def badIcon(self):
        """Returns a icon indicating a bad state.
        Returns:
            ImageTk PhotoImage"""
        if self.__class__.cbad_icon is None:
            self.__class__.cbad_icon = CTkImage(
                Image.open(getBadMarkIconPath()))
        return self.__class__.cbad_icon

    def waitingIcon(self):
        """Returns a icon indicating a waiting state.
        Returns:
            ImageTk PhotoImage"""
        if self.__class__.cwaiting_icon is None:
            self.__class__.cwaiting_icon = CTkImage(
                Image.open(getWaitingMarkIconPath()))
        return self.__class__.cwaiting_icon

    def __init__(self, parent, displayMsg="Connect to api"):
        """
        Open a child dialog of a tkinter application to connect to a pollenisator server.

        Args:
            parent: the tkinter parent view to use for this window construction.
            displayMsg: The message that will explain to the user what the form is.
        """
        self.parent = parent
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.title("Connect to server")
        self.app.resizable(False, False)
        appFrame = CTkFrame(self.app)
        self.app.bind("<Escape>", self.onError)
        self.rvalue = None
        self.parent = parent
        self.clientCfg = loadClientConfig()
        lbl = CTkLabel(self.app, text=displayMsg)
        lbl.pack()
        settings = Settings()
        settings.reloadLocalSettings()
        prev_hosts = settings.local_settings.get("hosts", [])
        row = 0
        if len(prev_hosts) > 0:
            CTkLabel(appFrame, text="Previous connections").grid(row=0, column=0, sticky="e")
            box_uri = CTkComboBox(appFrame, values=[x["url"] for x in prev_hosts], state="readonly", width=300, command=self.fill_with_uri)
            box_uri.grid(sticky="w", padx=5, row=0, column=1)
            box_uri.set(prev_hosts[0]["url"])
            row += 1
        lbl_hostname = CTkLabel(appFrame, text="Host : ")
        lbl_hostname.grid(row=row, column=0, sticky="e")
        self.ent_hostname = PopoEntry(
            appFrame, placeholder_text="127.0.0.1", validate="focusout", width=300, validatecommand=self.validateHost)
        self.ent_hostname.insert(tk.END, self.clientCfg["host"])
        self.ent_hostname.bind('<Return>', self.validateHost)
        self.ent_hostname.bind('<KP_Enter>', self.validateHost)
        self.ent_hostname.grid(row=row, column=1, sticky="w", padx=5)
        row += 1
        lbl_port = CTkLabel(appFrame, text="Port : ")
        lbl_port.grid(row=row, column=0, sticky="e")
        self.ent_port = PopoEntry(
            appFrame, placeholder_text="5000", validate="focusout", validatecommand=self.validateHost)
        self.ent_port.insert(tk.END, self.clientCfg.get("port", 5000), )
        self.ent_port.bind('<Return>', self.validateHost)
        self.ent_port.bind('<KP_Enter>', self.validateHost)
        self.ent_port.grid(row=row, column=1, sticky="w", padx=5)
        self.img_indicator = CTkLabel(appFrame, text="",image=self.waitingIcon())
        self.img_indicator.grid(row=row, column=2)
        row += 1

        self.var_https = tk.IntVar()
        lbl_https = CTkLabel(appFrame, text="https: ")
        lbl_https.grid(row=row, column=0, sticky="e")
        self.check_https = CTkSwitch(appFrame, variable=self.var_https, text="", onvalue=True, offvalue=False, command=self.validateHost)
        self.check_https.grid(row=row, column=1, sticky="w", padx=5)
        row += 1
        lbl_login = CTkLabel(appFrame, text="Login: ")
        lbl_login.grid(row=row, column=0, sticky="e")
        self.ent_login = PopoEntry(
            appFrame, placeholder_text="login", width=300)
        self.ent_login.grid(row=row, column=1, sticky="w", padx=5)
        row += 1
        lbl_passwd = CTkLabel(appFrame, text="Password: ")
        lbl_passwd.grid(row=row, column=0, sticky="e")
        self.password = tk.StringVar() 
        self.ent_passwd = PopoEntry(
            appFrame, placeholder_text="password", show="*", width=300, textvariable = self.password)
        self.ent_passwd.bind('<Return>', self.onOk)
        self.ent_passwd.grid(row=row, column=1, sticky="w", padx=5)
        appFrame.pack(padx=10, pady=10)
        self.ent_login.focus_set()
        cf1 = CollapsibleFrame(appFrame, text = "Advanced options", interior_padx=5, interior_pady=15)
        lbl_proxy = CTkLabel(cf1.interior, text="Proxy url : ")
        lbl_proxy.grid(row=0, column=0, sticky="e")
        self.ent_proxy = PopoEntry(cf1.interior, placeholder_text="http://127.0.0.1:8080")
        proxies = self.clientCfg.get("proxies", "")
        if proxies != "" and str(proxies).strip() != "{}":
            self.ent_proxy.insert(tk.END, proxies)
            cf1.toggle()
        self.ent_proxy.grid(row=0, column=1, sticky="w", padx=5)
        self.validateHost()

        cf1.update_width()

        cf1.grid(row=6,column=0, columnspan=2)

        self.ok_button = CTkButton(self.app, text="OK", command=self.onOk)
        self.ok_button.bind('<Return>', self.onOk)
        self.ok_button.pack(pady=10)
        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.focus_force()
            self.app.grab_set()
            self.app.lift()

        except tk.TclError:
            pass
        appFrame.tkraise()
        self.ent_login.focus_set()
        appFrame.tkraise()
        
    def getForm(self):
        """Return the content of this form
        Returns:
            a dict with values: host, mongo_port, sftp_port, ssl (string with value True or False),
                                user, password, sftp_user, sftp_password"""
        config = {}
        config["host"] = self.ent_hostname.get()
        config["port"] = self.ent_port.get()
        config["https"] = self.var_https.get()
        config["proxies"] = self.ent_proxy.get()
        return config


    def validateHost(self, _event=None):
        """Validate host on both mongo and sftp connections. Change icons on the dialog accordingly.
        Returns:
            - True if the server is reachable on both mongo and sftp services, False otherwise. Does not test authentication.
        """
        apiclient = APIClient.getInstance()
        apiclient.reinitConnection()
        config = self.getForm()
        self.img_indicator.configure(image=self.waitingIcon())
        res = apiclient.tryConnection(config)
        if res:
            self.img_indicator.configure(image=self.validIcon())
        else:
            self.img_indicator.configure(image=self.badIcon())
        return res

    def valideLogin(self):
        pass

    def fill_with_uri(self, uri):
        regex_host = r"^(http|https)://([^:]+):(\d+)"
        groups = re.search(regex_host,uri)
        if groups:
            self.ent_hostname.delete(0, tk.END)
            self.ent_hostname.insert(0, groups.group(2))
            self.ent_port.delete(0, tk.END)
            self.ent_port.insert(0, groups.group(3))
            self.var_https.set(groups.group(1) == "https")

    def onError(self, _event=None):
        self.rvalue = None
        self.app.quit()
        
    def onOk(self, event=None):
        """
        Called when the user clicked the validation button.
        Try a full connection with authentication to the host given.
        Side effects:
            - Open dialogs if the connection failed. Does not close this dialog.
            - If the connections succeeded : write the client.cfg file accordingly.
        """
        # send the data to the parent
        config = self.getForm()
        apiclient = APIClient.getInstance()
        apiclient.reinitConnection()
        res = apiclient.tryConnection(config)
        self.rvalue = False
        if res:
            #  pylint: disable=len-as-condition
            loginRes = apiclient.login(self.ent_login.get(), self.password.get())
            if loginRes:
                self.rvalue = True
                self.app.destroy()
            else:
                tk.messagebox.showerror("Connection failure", "The login/password you entered does not exists")
        else:
            tk.messagebox.showerror("Connection failure", "The host is not responding. Check if server is alive or if you have a local proxy configured.")



"""Ask the user to select a file or directory and then parse it with the selected parser"""
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
import tkinterDnD
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.forms.formpanel import FormPanel
import pollenisatorgui.core.components.utils as utils


class ChildDialogAskFile(CTk, tkinterDnD.tk.DnDWrapper):
    """
    Open a child dialog of a tkinter application to ask details about
    existing files parsing.
    """
    def _init_tkdnd(master: tk.Tk) -> None: #HACK to make work tkdnd with CTk
        """Add the tkdnd package to the auto_path, and import it"""
        #HACK Copied from directory with a package_dir updated
        platform = master.tk.call("tk", "windowingsystem")

        if platform == "win32":
            folder = "windows"
        elif platform == "x11":
            folder = "linux"
        elif platform == "aqua":
            folder = "mac"
        package_dir = os.path.join(os.path.dirname(os.path.abspath(tkinterDnD.tk.__file__)), folder)
        master.tk.call('lappend', 'auto_path', package_dir)
        TkDnDVersion = master.tk.call('package', 'require', 'tkdnd')
        return TkDnDVersion


    def __init__(self, parent, info="Choose a file", default_path=""):
        """
        Open a child dialog of a tkinter application to ask details about
        existing files parsing.

        Args:
            default_path: a default path to be added
        """
        super().__init__()
        self.TkDnDVersion = self._init_tkdnd()  #HACK to make work tkdnd with CTk
        self.settings = Settings()
        utils.setStyle(self, self.settings.local_settings.get("dark_mode", False))
        self.title(info)
        self.bind("<Escape>", self.onError)
        self.rvalue = None
        self.default = default_path
        appFrame = CTkFrame(self)
        self.form = FormPanel()
        self.form.addFormLabel(
            "Choose one file", info, side=tk.TOP)
        self.fileForm = self.form.addFormFile("File", ".+", '', width=50,
                              side=tk.TOP, mode="file")
        self.button = self.form.addFormButton("Cancel", self.onError, side=tk.RIGHT, 
                               fg_color=utils.getBackgroundColor(), text_color=utils.getTextColor(),
                               border_width=1, border_color="firebrick1", hover_color="tomato")
        self.form.addFormButton("OK", self.onOk, side=tk.RIGHT)

        self.form.constructView(appFrame)
        appFrame.pack(ipadx=10, ipady=10)

        try:
            self.wait_visibility()
            self.focus_force()
            self.grab_set()
            self.lift()
        except tk.TclError:
            pass
        self.mainloop()
    def quit(self):
        super().quit()
        self.destroy()

    def onError(self, _event=None):
        self.rvalue = None
        self.quit()

    def onOk(self, _event=None):
        """
        Called when the user clicked the validation button.
        launch parsing with selected parser on selected file/directory.
        Close the window.

        Args:
            _event: not used but mandatory
        """
        res, msg = self.form.checkForm()
        if not res:
            tk.messagebox.showwarning(
                "Form not validated", msg, parent=self)
            return
        files_paths = self.fileForm.getValue()
        try:
            filepath = files_paths[0]
        except IndexError:
            self.rvalue = None
            tk.messagebox.showwarning(
                "Form not validated", "At least one file is asked", parent=self)
            return
        self.rvalue = filepath
        self.quit()
        return



"""Ask the user to select a file or directory and then parse it with the selected parser"""
import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
import os
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.forms.formpanel import FormPanel
from pollenisatorgui.core.views.viewelement import ViewElement
from pollenisatorgui.core.application.dialogs.ChildDialogProgress import ChildDialogProgress
import pollenisatorgui.core.components.utils as utils


class ChildDialogFileParser:
    """
    Open a child dialog of a tkinter application to ask details about
    existing files parsing.
    """

    def __init__(self, parent, default_path=""):
        """
        Open a child dialog of a tkinter application to ask details about
        existing files parsing.

        Args:
            default_path: a default path to be added
        """
        self.app = CTkToplevel(parent)
        self.app.attributes("-type", "dialog")
        self.app.title("Upload result file")
        self.rvalue = None
        self.default = default_path
        appFrame = CTkFrame(self.app)
        self.app.bind("<Escape>", self.onError)
        apiclient = APIClient.getInstance()
        self.form = FormPanel()
        self.form.addFormLabel(
            "Import one file or choose a directory", "", side=tk.TOP)
        self.form.addFormFile("File", ".+", self.default, width=50,
                              side=tk.TOP, mode="file|directory")
        self.form.addFormLabel("Plugins", side=tk.TOP)
        self.form.addFormCombo(
            "Plugin", ["auto-detect"]+[x["plugin"] for x in apiclient.getPlugins()], "auto-detect", side=tk.TOP)
        self.form.addFormButton("Parse", self.onOk, side=tk.RIGHT)

        self.form.constructView(appFrame)
        appFrame.pack(pady=10,padx=10)

        try:
            self.app.wait_visibility()
            self.app.transient(parent)
            self.app.focus_force()
            self.app.grab_set()
            self.app.lift()
        except tk.TclError:
            pass

    def onError(self, event=None):
        self.app.destroy()
        self.rvalue = None
        return None

    def onOk(self, _event=None):
        """
        Called when the user clicked the validation button.
        launch parsing with selected parser on selected file/directory.
        Close the window.

        Args:
            _event: not used but mandatory
        """
        res, msg = self.form.checkForm()
        if not res:
            tk.messagebox.showwarning(
                "Form not validated", msg, parent=self.app)
            return
        notes = None
        tags = None
        form_values = self.form.getValue()
        form_values_as_dicts = ViewElement.list_tuple_to_dict(form_values)
        files_paths = form_values_as_dicts["File"]
        plugin = form_values_as_dicts["Plugin"]
        files = set()
        for filepath in files_paths:
            if os.path.isdir(filepath):
                # r=root, d=directories, f = files
                for r, _d, f in os.walk(filepath):
                    for fil in f:
                        files.add(os.path.join(r, fil))
            else:
                files.add(filepath)
        if not files:
            return
        dialog = ChildDialogProgress(self.app, "Importing files", "Importing "+str(
            len(files)) + " files. Please wait for a few seconds.", 100.0/len(files), "determinate")
        dialog.show(len(files))
        # LOOP ON FOLDER FILES
        results = {}
        apiclient = APIClient.getInstance()
        for f_i, file_path in enumerate(files):
            additional_results = apiclient.importExistingResultFile(file_path, plugin)
            for key, val in additional_results.items():
                results[key] = results.get(key, 0) + val
            dialog.update()
        dialog.destroy()
        # DISPLAY RESULTS
        presResults = ""
        filesIgnored = 0
        for key, value in results.items():
            presResults += str(value) + " " + str(key)+".\n"
            if key == "Ignored":
                filesIgnored += 1
        if plugin == "auto-detect":
            if filesIgnored > 0:
                tk.messagebox.showwarning(
                    "Auto-detect ended", presResults, parent=self.app)
            else:
                tk.messagebox.showinfo("Auto-detect ended", presResults, parent=self.app)
        else:
            if filesIgnored > 0:
                tk.messagebox.showwarning(
                    "Parsing ended", presResults, parent=self.app)
            else:
                tk.messagebox.showinfo("Parsing ended", presResults, parent=self.app)

        self.rvalue = None
        self.app.quit()



"""View for ip list object. Present an multi insertion form to user when interacted with."""

from pollenisatorgui.core.views.viewelement import ViewElement


class MultipleIpView(ViewElement):
    """View for ip list object. Present an multi insertion form to user when interacted with."""


    def openInsertWindow(self):
        """
        Creates a tkinter form using Forms classes. This form aims to insert many new Ips
        """
        toppanel = self.form.addFormPanel(grid=True)
        toppanel.addFormLabel("IPs")
        toppanel.addFormHelper(
            "Add IP or hostnames that you want to include in the scans\nEach ip/hostname must on a separated line", column=1)
        self.form.addFormText("IPs", r"^(?!https?://).+$",
                              "", None, side="top")
        self.completeInsertWindow()

    def getAdditionalContextualCommands(self):
        return {"Insert Ips/Hosts":self.openInsertWindow}


from bson import ObjectId
from pollenisatorgui.core.application.dialogs.ChildDialogAskFile import ChildDialogAskFile
from pollenisatorgui.core.components.apiclient import APIClient
import pollenisatorgui.core.components.utils as utils
import os


def main(apiclient, appli, **kwargs):
    print(kwargs)
    john_path = utils.which_expand_alias("john")
    if not john_path:
        return False, "binary 'john' is not in the PATH."
    APIClient.setInstance(apiclient)
    secrets = None
    if kwargs.get("target_type").lower() == "user" or kwargs.get("target_type").lower() == "computer":
        info =  apiclient.find(kwargs.get("target_type").lower()+"s", {"type":kwargs.get("target_type").lower(), "_id":ObjectId(kwargs.get("target_iid", ""))}, False)
        if info is not None:
            secrets = info.get("infos", {}).get("secrets", [])
    if secrets is None:
        dialog = ChildDialogAskFile(None, "Hashes to crack:")
        if dialog.rvalue is None:
            return False, "No hash list given"
        hash_file_name = dialog.rvalue
        if not os.path.exists(hash_file_name):
            return False, "hash file list given not found."
    else:
        export_dir = utils.getExportDir()
        out_path = os.path.join(export_dir, str(kwargs.get("target_iid")))
        try:
            os.makedirs(out_path)
        except:
            pass
        hash_file_name = os.path.join(out_path, "secrets.txt")
        with open(hash_file_name, "w") as f:
            f.write("\n".join(secrets))
    dialog = ChildDialogAskFile(None, "wordlist to use:")
    
    if dialog.rvalue is None:
        return False, "No wordlist list given"
    wordlist_file_name = dialog.rvalue
    if not os.path.exists(wordlist_file_name):
        return False, "wordlist file list given not found."
    iid = appli.launch_in_terminal(kwargs.get("default_target",None), "Crack hashes", f"{john_path} --wordlist={wordlist_file_name} {hash_file_name}")
    iid = appli.launch_in_terminal(kwargs.get("default_target",None), "Result hashes", f"{john_path} --show {hash_file_name}")
    return True, f"Launched john in terminal, if it fails, try another wordlist or add rules."



"""View for checkitem object. Handle node in treeview and present forms to user when interacted with."""

from pollenisatorgui.core.application.dialogs.ChildDialogDefectView import ChildDialogDefectView
from pollenisatorgui.core.components.apiclient import APIClient
from pollenisatorgui.core.controllers.checkinstancecontroller import CheckInstanceController
from pollenisatorgui.core.controllers.commandcontroller import CommandController
from pollenisatorgui.core.models.defect import Defect
from pollenisatorgui.core.views.checkinstanceview import CheckInstanceView
from pollenisatorgui.core.views.commandview import CommandView
from pollenisatorgui.core.views.viewelement import ViewElement
from pollenisatorgui.core.views.multitodocheckinstanceview import MultiTodoCheckInstanceView
from pollenisatorgui.core.components.settings import Settings
from pollenisatorgui.core.models.command import Command
from pollenisatorgui.core.controllers.checkitemcontroller import CheckItemController
from pollenisatorgui.core.models.checkitem import CheckItem
from pollenisatorgui.core.components.scriptmanager import ScriptManager
import pollenisatorgui.core.components.utils as utils
from customtkinter import *
from bson import ObjectId
import tkinter as tk

class CheckItemView(ViewElement):
    """
    View for CheckItemView object. Handle node in treeview and present forms to user when interacted with.
    Attributes:
        icon: icon name to show in treeview. Icon filename must be in icon directory.
    """
    default_icon = 'checklist.png'
    cached_default_icon = None
    icon = "checklist.png"
    icon_auto = 'auto.png'
    cached_icon_auto = None

    def __init__(self, appTw, appViewFrame, mainApp, controller):
        """Constructor
        Args:
            appTw: a PollenisatorTreeview instance to put this view in
            appViewFrame: an view frame to build the forms in.
            mainApp: the Application instance
            controller: a CommandController for this view.
        """
        self.menuContextuel = None
        self.widgetMenuOpen = None
        self.checktypeForm = None
        self._limit_childrens = 25
        super().__init__(appTw, appViewFrame, mainApp, controller)

    def getIcon(self):
        """
        Load the object icon in cache if it is not yet done, and returns it

        Return:
            Returns the icon representing this object.
        """
        isAuto = self.controller.isAuto()
        if isAuto: # order is important as "done" is not_done but not the other way
            ui = self.__class__.icon_auto
            cache = self.__class__.cached_icon_auto
            iconStatus = "auto"
        else:
            cache = self.__class__.cached_default_icon
            ui = self.__class__.default_icon
            iconStatus = "default"
        if cache is None:
            from PIL import Image, ImageTk

            path = utils.getIcon(ui)
            if iconStatus == "auto":
                self.__class__.cached_icon_auto = ImageTk.PhotoImage(
                    Image.open(path))
                return self.__class__.cached_icon_auto
            else:
                self.__class__.cached_default_icon = ImageTk.PhotoImage(
                    Image.open(path))
                return self.__class__.cached_default_icon
        return cache

    def _commonWindowForms(self, default={}, action="modify"):
        """Construct form parts identical between Modify and Insert forms
        Args:
            default: a dict of default values for inputs (priority, priority, max_thread). Default to empty respectively "0", "0", "1"
        """
        self.form.addFormHidden("Step", default.get("step", 0))
        self.form.addFormHidden("Parent", default.get("parent", ""))
        panel_bottom = self.form.addFormPanel(grid=True)
        panel_bottom.addFormLabel("Check type", row=0, column=0)
        self.checktypeForm = panel_bottom.addFormCombo("Check type", ("manual_commands", "auto_commands", "script", "manual"), 
                                                        default=default.get("check_type","manual"),  
                                                        command=lambda ev: self.updateCheckType(action),
                                                        row=0, column=1)
        panel_bottom.addFormLabel("Trigger", row=0, column=0)
        panel_bottom.addFormChecklist("Pentest types", list(Settings.getPentestTypes().keys()), default.get("pentest_types", []), row=1, column=1)
        panel_bottom = self.form.addFormPanel(grid=True)
        apiclient = APIClient.getInstance()
        lvls = apiclient.getTriggerLevels()
        panel_bottom.addFormLabel("Trigger", row=0, column=0)
        true_lvl = default.get("lvl", "port:onServiceUpdate")
        if true_lvl.startswith("tag:"):
            true_lvl = ":".join(true_lvl.split(":")[:2])+":str"
        self.triggerLevelForm = panel_bottom.addFormCombo(
            "Level", lvls, true_lvl,  command=lambda ev: self.triggerLevelUpdate(true_lvl), width=200, row=0, column=1)
        panel_bottom.addFormHelper(
            "When the event is triggered, \na instance for this check will be created on the triggering object", row=0, column=2)
        self.panel_trigger_options = self.form.addFormPanel(grid=True)
        if default.get("lvl","port:onServiceUpdate").startswith("port"):
            self.panel_trigger_options.clear()
            self.panel_trigger_options.addFormLabel("Ports/Services (if level is port only)")
            self.panel_trigger_options.addFormStr(
                "Ports/Services", r"^(\d{1,5}|[^\,]+)?(?:,(\d{1,5}|[^\,]+))*$", default.get("ports", ""), column=1)
            self.panel_trigger_options.addFormHelper(
                "Services, ports or port ranges.\nthis list must be separated by a comma, if no protocol is specified, tcp/ will be used.\n Example: ssl/http,https,http/ssl,0-65535,443...", column=2)
        if default.get("lvl", "").startswith("tag:"):
            lvl_split = default.get("lvl", "").split(":")
            if len(lvl_split) == 3:
                tag_name = lvl_split[2] 
            else:
                tag_name = ""
            self.panel_trigger_options.clear()
            self.panel_trigger_options.addFormLabel("Tag name")
            self.panel_trigger_options.addFormStr(
                "Tag Name", r".+", tag_name, column=1)
            self.panel_trigger_options.addFormHelper(
                "Tag filter to trigger this check", column=2)
        panel_bottom = self.form.addFormPanel(grid=True)
        panel_bottom.addFormLabel("Category", row=1, column=0)
        panel_bottom.addFormStr("Category", r"", default.get("category", ""), row=1, column=1)
        panel = self.form.addFormPanel(grid=True)
        panel.addFormLabel("Priority")
        panel.addFormStr(
            "Priority", r"\d+", default["priority"], width=40, column=1)
        panel.addFormHelper(
            "Defines the priority of this group of command when an auto scan is running.\nAutoscan will try to launch the highest priority (0 is max) and the highest+1.", column=2)
        
        panel_bottom = self.form.addFormPanel()
        panel_bottom.addFormLabel("Description")
        panel_bottom.addFormText("Description", r"", default.get("description", ""), side="right")
        if "commands" in default.get("check_type", "manual"):
            formCommands = self.form.addFormPanel(grid=True)
            formCommands.addFormLabel("Shared threads")
            formCommands.addFormStr("Shared threads", r"\d+",
                            default["max_thread"], width=40, column=1)
            panel.addFormHelper(
                "Number of parallel execution allowed for every command in this group at any given moment.", column=2)
            formTv = self.form.addFormPanel(side=tk.TOP, fill=tk.X, pady=5)
            formSearchCommand = formTv.addFormPanel(side=tk.LEFT, fill=tk.Y, pady=5, padx=5)
            formSearchCommand.addFormSearchBar("Commands search", self.searchCallback, self.form, side=tk.TOP)
            formSearchCommand.addFormSeparator()
            formSearchCommand.addFormButton("Create new command", self.create_command_callback, side=tk.TOP)
            commands = default.get("commands")
            if commands is None:
                tv_commands = ["", ""]
            else:
                tv_commands = []
                for command in commands:
                    comm = Command.fetchObject({"_id":ObjectId(command)})
                    if comm is not None:
                        tv_commands.append((comm.name, str(command)))
            self.treeview_commands = formTv.addFormTreevw(
                "Commands", ("Command names",), tv_commands, doubleClickBinds=[self.onCommandDoubleClick],height=5, width=30, pady=5, fill=tk.X, side=tk.RIGHT)
        elif "script" in default.get("check_type", "manual"):
            formTv = self.form.addFormPanel(side=tk.TOP, fill=tk.X, pady=5)
            formTv.addFormLabel("Script", side=tk.LEFT)
            self.textForm = formTv.addFormStr("Script", ".+", default.get("script", ""))
            btn = formTv.addFormButton("Browse", lambda _event : self.browseScriptCallback(self.textForm))
        ### form defects
        self.formDefects = self.form.addFormPanel(side=tk.TOP, fill=tk.X, pady=5)
        formSearchDefects = self.formDefects.addFormPanel(side=tk.LEFT, fill=tk.Y, pady=5, padx=5)
        plugins = apiclient.getPlugins()
        tags = set()
        for plugin in plugins:
            tags = tags.union(set([tag.get("name") for tag in plugin.get("tags", [])]))
        self.possible_tags = sorted(list(tags))
        formSearchDefects.addFormSearchBar("Defect search", self.searchDefectCallback, self.formDefects, side=tk.TOP)
        formSearchDefects.addFormSeparator()
        formSearchDefects.addFormButton("Create new defect template", self.create_defect_callback, side=tk.TOP)
        defect_tags = default.get("defect_tags")
        if defect_tags is None:
            tv_defects = ["", "", ""]
        else:
            tv_defects = []
            for tag, defect in defect_tags:
                try:
                    defect = ObjectId(defect)
                except:
                    continue
                defect_o = Defect.getTemplateById(ObjectId(defect))
                if defect_o is not None:
                    tv_defects.append((defect_o.title,  tag, str(defect_o.getId())))
        self.treeview_defects = self.formDefects.addFormTreevw(
            "Defects", ("Defect names", "Tag association"), tv_defects, doubleClickBinds=[self.onDefectDoubleClick, self.possible_tags],height=5, width=30, pady=5, fill=tk.X, side=tk.RIGHT)
        
    def onCommandDoubleClick(self, oldval):
        command_o = Command.fetchObject({"name":oldval})
        if command_o is None:
            return
        view = CommandView(self.appliTw, self.appliViewFrame, self.mainApp, CommandController(command_o))
        view.openInDialog()

    def onDefectDoubleClick(self, oldval):
        defect_o = Defect.fetchObject({"title":oldval})
        if defect_o is None:
            return
        dialog = ChildDialogDefectView(self.mainApp, "Edit defect", self.settings, defect_o)
    
    def reopen(self):
        if self.is_insert_view:
            self.openInsertWindow()
            return
        self.openModifyWindow()

    def updateCheckType(self, action):
        form_values = self.form.getValue()
        form_values_as_dicts = ViewElement.list_tuple_to_dict(form_values)
        self.controller.doUpdate(form_values_as_dicts, updateInDb=False)
        self.form.clear()
        self.reopen()

    def triggerLevelUpdate(self, old_value):
        if self.triggerLevelForm.getValue() != old_value:
            form_values = self.form.getValue()
            form_values_as_dicts = ViewElement.list_tuple_to_dict(form_values)
            self.controller.doUpdate(form_values_as_dicts, updateInDb=False)
            self.form.clear()
            self.reopen()

    def browseScriptCallback(self, textForm):
        scriptManagerInst = ScriptManager()
        scriptManagerInst.initUI(self.mainApp, whatfor="select")
        self.mainApp.wait_window(scriptManagerInst.app)
        liste = scriptManagerInst.rvalue
        if liste is not None:
            textForm.setValue(", ".join(liste))
        
    def openModifyWindow(self, **kwargs):
        """
        Creates a tkinter form using Forms classes. This form aims to update or delete an existing Command
        """
        if kwargs.get("treevw") == "cheatsheet":
            self.openModifyWindowFromCheasheet()
            return
        else:
            self.openModifyWindowFromPentestView()
            return
        
    def openModifyWindowFromPentestView(self):
        self.is_insert_view = False
        self.completeModifyWindow(editable=False, addTags=False)

    def openModifyWindowFromCheasheet(self):
        self.is_insert_view = False
        self.form.clear()
        modelData = self.controller.getData()
        self._initContextualMenu()
        panel_top = self.form.addFormPanel(grid=True)
        panel_top.addFormLabel("Title", column=0)
        panel_top.addFormStr("Title", r".+", default=modelData.get("title", ""), column=1)

        self._commonWindowForms(modelData, action="modify")
        #if modelData.get("parent") is None:
            #self.form.addFormButton("Add a step", self.addStep)
        self.completeModifyWindow(editable=True, addTags=False)



    def addStep(self, event=None):
        newCheckItem = CheckItem({"parent":str(self.controller.getDbId())})
        view = CheckItemView(self.appliTw, self.appliViewFrame, self.mainApp, CheckItemController(newCheckItem))
        view.openInsertWindow()

    def openInsertWindow(self):
        """
        Creates a tkinter form using Forms classes. This form aims to insert a new Command
        """
        self.is_insert_view = True
        self.form.clear()
        data = self.controller.getData()
        self._initContextualMenu()
        panel_top = self.form.addFormPanel(grid=True)
        panel_top.addFormLabel("Title", column=0)
        panel_top.addFormStr("Title", r".+", default=data.get("title", ""), column=1)
        
       
        self._commonWindowForms(data, action="insert")
        self.completeInsertWindow()

    def _insertChildren(self):
        checks = self.controller.getChecks()
        if len(checks) > self._limit_childrens:
            checks_done = {}
            checks_running = {}
            checks_not_done = {}
            for check in checks:
                if check.getStatus() == "done":
                    checks_done[check.getId()] = check
                elif check.getStatus() == "running":
                    checks_running[check.getId()] = check
                else:
                    checks_not_done[check.getId()] = check
            title = str(checks[0])
            #self.appliTw.insert(self.parent, 'end', text=f"{title} (todo {len(self.checks)})", image=CheckInstanceView.getIcon({"status":"todo"}))
            #self.appliTw.insert(self.parent, 'end', text=f"{title}(running {len(self.checks)})",  image=CheckInstanceView.getIcon({"status":"running"}))
            #self.appliTw.insert(self.parent, 'end', text=f"{title}(done {len(self.checks)})",  image=CheckInstanceView.getIcon({"status":"done"}))
            checks_vw = MultiTodoCheckInstanceView(self.appliTw, self.appliViewFrame, self.mainApp, checks_not_done, self.controller.getDbId())
            checks_vw.addInTreeview(title)
            return
        
        for check in checks:
            check_o = CheckInstanceController(check)
            check_vw = CheckInstanceView(self.appliTw, self.appliViewFrame, self.mainApp, check_o)
            check_vw.addInTreeview(str(self.controller.getDbId()), addChildren=False)


    def addInTreeview(self, parentNode=None, **kwargs):
        """Add this view in treeview. Also stores infos in application treeview.
        Args:
            parentNode: if None, will calculate the parent. If setted, forces the node to be inserted inside given parentNode.
        """
        with_category = kwargs.get("with_category", False)
        addChildren = kwargs.get("addChildren", False)
        count_children = kwargs.get("count_children", -1)
        self.appliTw.views[str(self.controller.getDbId())] = {"view": self}
        if parentNode is None:
            parentNode = self.getParentNode(with_category)
        try:
            text = str(self.controller.getModelRepr())
            if count_children > 0:
                text += f" ({count_children})"
            #FASTER THAN self.appliTw.insert(parentNode, "end", str(
            #    self.controller.getDbId()), text=str(self.controller.getModelRepr()), tags=self.controller.getTags(), image=self.getIcon())
            node = self.appliTw.tk.call(self.appliTw._w, "insert", parentNode, "end", "-id", str(self.controller.getDbId()), 
                                 "-text", text, "-image", self.getIcon())
        except tk.TclError as e:
            pass
        if not addChildren and getattr(self.appliTw, "lazyload", False) and not self.mainApp.searchMode:
            try:
                self.appliTw.tk.call(self.appliTw._w, "insert", str(self.controller.getDbId()), "end", "-id", str(self.controller.getDbId())+"|<Empty>", 
                                 "-text","<Empty>")
            except tk.TclError as e:
                pass
        elif addChildren:
            self._insertChildren()
        if hasattr(self.appliTw, "hide"):
            if not self.mainApp.settings.is_checklist_view():
                self.hide("checklist_view")
            if self.mainApp.settings.is_show_only_manual() and self.controller.isAuto():
                self.hide("filter_manual")
            

    def getParentNode(self, with_category=False):
        """
        Return the id of the parent node in treeview.

        Returns:
            return the saved command_node node inside the Appli class.
        """
        if with_category:
            category = self.controller.getCategory()
            try:
                self.appliTw.insert("", "end", category, text=self.controller.getCategory(),  image=self.__class__.getClassIcon())
            except tk.TclError:
                pass
            if hasattr(self.appliTw, "hide") and not self.mainApp.settings.is_checklist_view():
                self.appliTw.hide(category, "checklist_view")
            return category
        parent = self.controller.getParent()
        if parent is not None and parent != "":
            return parent
        
        return "" # category

    def _initContextualMenu(self):
        """Initiate contextual menu with variables"""
        self.menuContextuel = utils.craftMenuWithStyle(self.appliViewFrame)

    def popup(self, event):
        """
        Fill the self.widgetMenuOpen and reraise the event in the editing window contextual menu

        Args:
            event: a ttk Treeview event autofilled. Contains information on what treeview node was clicked.
        """
        self.widgetMenuOpen = event.widget
        self.menuContextuel.tk_popup(event.x_root, event.y_root)
        self.menuContextuel.focus_set()
        self.menuContextuel.bind('<FocusOut>', self.popupFocusOut)

    def popupFocusOut(self, _event=None):
        """
        Called when the contextual menu is unfocused
        Args:
            _event: a ttk event autofilled. not used but mandatory.
        """
        self.menuContextuel.unpost()

    def searchDefectCallback(self, searchreq):
        ret = []
        defects_obj, defects_errors = APIClient.getInstance().searchDefect(searchreq)#, check_api=True
        if defects_obj:
            for i, defect in enumerate(defects_obj):
                ret.append({"TITLE": defect["title"], "defects":{"text":defect["title"], "values":("", str(defect["_id"]))}})
        else:
            return [], defects_errors
        return ret, ""

    def searchCallback(self, searchreq):
        apiclient = APIClient.getInstance()
        commands = apiclient.findInDb("pollenisator", "commands", {"name":{'$regex':searchreq}})
        if commands is None:
            return [], "Invalid response from API"

        ret = [{"TITLE":command["name"], "commands":{"text":command["name"], "values":(0, str(command["_id"]))}} for command in commands]
        return ret, ""
    
    def create_command_callback(self, _event=None):
        view = CommandView(self.appliTw, self.appliViewFrame, self.mainApp, CommandController(Command()))
        result = view.openInDialog(is_insert=True)
        if result is not None:
            result, msg = result # unpack if not none
            comm = Command.fetchObject({"_id":ObjectId(msg)})
            if comm is not None:
                self.treeview_commands.addItem("", "end", str(msg), text=comm.name, values=(0, str(comm.getId()),))
            else:
                tk.messagebox.showerror("Error inserting command", msg)

    def create_defect_callback(self, _event=None):
        dialog = ChildDialogDefectView(self.mainApp, "Add a security defect template", self.mainApp.settings, as_template=True)
        try:
            self.mainApp.wait_window(dialog.app)	
        except tk.TclError:
            pass
        if dialog.rvalue:
            res, msg = dialog.rvalue
            defect_o = Defect.fetchObject({"_id":ObjectId(msg)})
            if defect_o is not None:
                self.treeview_defects.addItem("", "end", str(msg), text=defect_o.name, values=(0, str(defect_o.getId()),))
            else:
                tk.messagebox.showerror("Error inserting defect template", msg)


    def updateReceived(self, obj=None, old_obj=None):
        """Called when a command update is received by notification.
        Update the command treeview item (resulting in icon reloading)
        """
        # if self.controller.isMine():
        #     try:
        #         self.appliTw.move(str(self.controller.model.getId()), self.appliTw.my_commands_node, "end")
        #     except tk.TclError:
        #         print("WARNING: Update received for a non existing command "+str(self.controller.getModelRepr()))
        # else:
        #     try:
        #         self.appliTw.move(str(self.controller.model.getId()), self.appliTw.commands_node, "end")
        #     except tk.TclError:
        #         print("WARNING: Update received for a non existing command "+str(self.controller.getModelRepr()))
        
        super().updateReceived()

    def key(self):
        """Returns a key for sorting this node
        Returns:
            tuple, key to sort
        """
        return tuple([ord(c) for c in str(self.controller.getModelRepr()).lower()])

    # function that converts a string into a lsit of ascii values
    def convert(self, string):  
        li = list(string.split(" ")) 
        return li   



"""Describe tkinter combobox with default common args"""

import tkinter as tk
import tkinter.ttk as ttk
from customtkinter import *
from pollenisatorgui.core.forms.form import Form


class FormCombo(Form):
    """
    Form field representing a combobox.
    Default setted values: 
        state="readonly"
        if pack : padx = pady = 5, side = "right"
        if grid: row = column = 0 sticky = "west"
    Additional values to kwargs:
        binds:  a dictionnary of tkinter binding with shortcut as key and callback as value
    """

    def __init__(self, name, choicesList, default, **kwargs):
        """
        Constructor for a form checkbox

        Args:
            name: the checklist name (id).
            choicesList: a list of string forming all the possible choices.
            default: a list of string that should be prechecked if in the choice list.
            kwargs: same keyword args as you would give to CTkComboBox
        """
        super().__init__(name)
        self.choicesList = choicesList
        self.default = default
        self.kwargs = kwargs
        self.box = None
        self.wid_kwargs = None

    

    def constructView(self, parent):
        """
        Create the combobox view inside the parent view given

        Args:
            parent: parent FormPanel.
        """
        kw = {}
        width = self.getKw("width", None)
        if width is not None:
            kw["width"] = int(width)
        self.box = CTkComboBox(parent.panel, values=tuple(
            self.choicesList), state=self.getKw("state", "normal"), command=self.getKw("command",None), **kw)
        if self.default is not None:
            self.box.set(self.default)
        binds = self.getKw("binds", {})
        self.box.bind('<KeyRelease>', self.check_input)
        for bind in binds:
            self.box.bind(bind, binds[bind])
        if parent.gridLayout:
            self.box.grid(row=self.getKw("row", 0), column=self.getKw(
                "column", 0), sticky=self.getKw("sticky", tk.W))
        else:
            self.box.pack(side=self.getKw("side", "right"), padx=self.getKw(
                "padx", 10), pady=self.getKw("pady", 5), anchor=self.getKw("anchor", None))
        if self.wid_kwargs is not None:
            self.box.configure(**self.wid_kwargs)

    def getValue(self):
        """
        Return the form value. Required for a form.

        Returns:
            Return the selected text inside the comboxbox.
        """
        v = self.box.get()
        if v == "<Empty>":
            return ""
        return v

    def get(self):
        """alias of getValue"""
        return self.getValue()
    
    def setValue(self, newval):
        """
        Set the combo value.
        Args:
            newval: the new value to be set inside the combobox
        """
        if isinstance(newval, list):
            #self.choicesList = newval
            if newval:
                self.box.set(newval[0])
        elif newval not in self.choicesList:
            self.choicesList.append(newval)
        self.box.set(newval)
        command = self.box.cget("command")
        if command is not None:
            command()

    def set(self, newval):
        # alias for setValue
        return self.setValue(newval)

    def checkForm(self):
        """
        Check if this form is correctly filled. Formal verification if the selected value is still on the choice list.

        Returns:
            {
                "correct": True if the form is correctly filled, False otherwise.
                "msg": A message indicating what is not correctly filled.
            }
        """
        if "<Empty>" in self.choicesList and self.getValue() == "":
            return True, ""
        if self.getValue() not in self.choicesList:
            return False, self.name+" values ("+str(self.getValue())+") not in the accepted list."
        return True, ""

    def setFocus(self):
        """Set the focus to the ttk combobox.
        """
        self.box.focus_set()

    def configure(self, **kwargs):
        if "values" in kwargs:
            self.box.configure(values=kwargs["values"])
            self.choicesList = kwargs["values"]
            del kwargs["values"]
        self.wid_kwargs = kwargs

    def winfo_reqwidth(self):
        return self.box.winfo_reqwidth()
    
    def winfo_reqheight(self):
        return self.box.winfo_reqheight()
    
    def winfo_rootx(self):
        return self.box.winfo_rootx()
    
    def winfo_rooty(self):
        return self.box.winfo_rooty()
    
    def check_input(self, event):
        if event is not None:
            if event.keysym != "BackSpace" and event.keysym != "Delete" and (len(event.keysym) > 1 or not event.keysym.isalnum()):
                return
        value = event.widget.get()
        if value == '':
            self.box['values'] = self.choicesList
        else:
            data = []
            for item in self.choicesList:
                if isinstance(value, str):
                    if value.lower() in item.lower():
                        data.append(item)

            self.box['values'] = data
            if len(data) == 1:
                self.setValue(data)
                self.box.event_generate("<Down>")
